'''
Author: dbliu shaxunyeman@gmail.com
Date: 2024-09-02 17:55:19
LastEditors: dbliu shaxunyeman@gmail.com
LastEditTime: 2024-09-05 09:02:52
FilePath: /magic-doc/magic_doc/conv/pdf.py
Description: 
'''
from io import BytesIO

from loguru import logger
from werkzeug.datastructures import FileStorage

from magic_doc.contrib.pdf.pdf_extractor import PDFExtractor
from magic_doc.conv.base import BaseConv
from magic_doc.progress.filepupdator import FileBaseProgressUpdator
from magic_doc.progress.pupdator import ConvProgressUpdator
from magic_doc.conv.base import ParseFailed
from magic_doc.conv.pdf_pp_structurev2 import Pdf as liteOcr

class Pdf(BaseConv):
    def __init__(self, allowed_failure=True):
        self.allowed_failure = allowed_failure

    def to_md(self, bits: bytes | str, pupdator: ConvProgressUpdator, **kwargs) -> str:
        pdf_extractor = PDFExtractor()
        buf = BytesIO(bits)  # type: ignore
        content = pdf_extractor.run("stream io data", FileStorage(buf, "fake.pdf"))
        arr = []
        pupdator.update(0)

        N = len(content)
        progress_h = {N * i // 100: 1 for i in range(10, 100, 10)}
        for idx, page in enumerate(content):
            if idx in progress_h:
                pupdator.update(idx * 100 // N)
            for record in page.get("content_list", []):
                arr.append(record.get("data", ""))

        text_all = ""
        for content in arr:
             text_all += content
        def calculate_not_printable_rate(text):
             printable = sum(1 for c in text if c.isprintable())
             total = len(text)
             if total == 0:
                 return 0  # 避免除以零的错误
             return (total - printable) / total
        not_printable_rate = calculate_not_printable_rate(text_all)
        if not_printable_rate > 0.02:
            if self.allowed_failure:
                raise ParseFailed
            else:
                liteOcrPdf = liteOcr()
                return liteOcrPdf.to_md(bits, pupdator)
        else:
            pupdator.update(100)
            return "\n\n".join(arr)


if __name__ == "__main__":
    if 1:
        with open("/opt/data/pdf/20240423/pdf_test2/ol006018w.pdf", "rb") as f:
            bits_data = f.read()
            parser = Pdf()
            md_content = parser.to_md(bits_data, FileBaseProgressUpdator("debug/progress.txt"))

        with open("debug/pdf2md.md", "w") as f:
            f.write(md_content)



import os
from .visualizer import Visualizer
from .rcnn_vl import *
from .backbone import *

from detectron2.config import get_cfg
from detectron2.config import CfgNode as CN
from detectron2.data import MetadataCatalog, DatasetCatalog
from detectron2.data.datasets import register_coco_instances
from detectron2.engine import DefaultTrainer, default_argument_parser, default_setup, launch, DefaultPredictor
from magic_doc.utils import get_repo_directory

def add_vit_config(cfg):
    """
    Add config for VIT.
    """
    _C = cfg

    _C.MODEL.VIT = CN()

    # CoaT model name.
    _C.MODEL.VIT.NAME = ""

    # Output features from CoaT backbone.
    _C.MODEL.VIT.OUT_FEATURES = ["layer3", "layer5", "layer7", "layer11"]

    _C.MODEL.VIT.IMG_SIZE = [224, 224]

    _C.MODEL.VIT.POS_TYPE = "shared_rel"

    _C.MODEL.VIT.DROP_PATH = 0.

    _C.MODEL.VIT.MODEL_KWARGS = "{}"

    _C.SOLVER.OPTIMIZER = "ADAMW"

    _C.SOLVER.BACKBONE_MULTIPLIER = 1.0

    _C.AUG = CN()

    _C.AUG.DETR = False

    _C.MODEL.IMAGE_ONLY = True
    _C.PUBLAYNET_DATA_DIR_TRAIN = ""
    _C.PUBLAYNET_DATA_DIR_TEST = ""
    _C.FOOTNOTE_DATA_DIR_TRAIN = ""
    _C.FOOTNOTE_DATA_DIR_VAL = ""
    _C.SCIHUB_DATA_DIR_TRAIN = ""
    _C.SCIHUB_DATA_DIR_TEST = ""
    _C.JIAOCAI_DATA_DIR_TRAIN = ""
    _C.JIAOCAI_DATA_DIR_TEST = ""
    _C.ICDAR_DATA_DIR_TRAIN = ""
    _C.ICDAR_DATA_DIR_TEST = ""
    _C.M6DOC_DATA_DIR_TEST = ""
    _C.DOCSTRUCTBENCH_DATA_DIR_TEST = ""
    _C.DOCSTRUCTBENCHv2_DATA_DIR_TEST = ""
    _C.CACHE_DIR = ""
    _C.MODEL.CONFIG_PATH = ""

    # effective update steps would be MAX_ITER/GRADIENT_ACCUMULATION_STEPS
    # maybe need to set MAX_ITER *= GRADIENT_ACCUMULATION_STEPS
    _C.SOLVER.GRADIENT_ACCUMULATION_STEPS = 1


def setup(args, device):
    """
    Create configs and perform basic setups.
    """
    cfg = get_cfg()
    # add_coat_config(cfg)
    add_vit_config(cfg)
    cfg.merge_from_file(args.config_file)
    cfg.MODEL.ROI_HEADS.SCORE_THRESH_TEST = 0.2  # set threshold for this model
    cfg.merge_from_list(args.opts)

    # 使用统一的device配置
    cfg.MODEL.DEVICE = device

    cfg.freeze()
    default_setup(cfg, args)
    
    """ 
    #TODO: 可以去掉？
    register_coco_instances(
        "scihub_train",
        {},
        cfg.SCIHUB_DATA_DIR_TRAIN + ".json",
        cfg.SCIHUB_DATA_DIR_TRAIN
    )
    """
    
    return cfg


class DotDict(dict):
    def __init__(self, *args, **kwargs):
        super(DotDict, self).__init__(*args, **kwargs)

    def __getattr__(self, key):
        if key not in self.keys():
            return None
        value = self[key]
        if isinstance(value, dict):
            value = DotDict(value)
        return value
    
    def __setattr__(self, key, value):
        self[key] = value
        
class Layoutlmv3_Predictor(object):
    def __init__(self, weights, device='cuda'):
        layout_args = {
            "config_file": os.path.join(get_repo_directory(), "resources/model/layoutlmv3/layoutlmv3_base_inference.yaml"), # TODO 修改配置路径
            "resume": False,
            "eval_only": False,
            "num_gpus": 1,
            "num_machines": 1,
            "machine_rank": 0,
            "dist_url": "tcp://127.0.0.1:57823",
            "opts": ["MODEL.WEIGHTS", weights],
        }
        layout_args = DotDict(layout_args)
        
        cfg = setup(layout_args, device)
        self.mapping = ["title", "plain text", "abandon", "figure", "figure_caption", "table", "table_caption", "table_footnote", "isolate_formula", "formula_caption"]
        MetadataCatalog.get(cfg.DATASETS.TRAIN[0]).thing_classes = self.mapping
        self.predictor = DefaultPredictor(cfg)
        
    def __call__(self, image, ignore_catids=[]):
        page_layout_result = {
            "layout_dets": []
        }
        outputs = self.predictor(image)
        boxes = outputs["instances"].to("cpu")._fields["pred_boxes"].tensor.tolist()
        labels = outputs["instances"].to("cpu")._fields["pred_classes"].tolist()
        scores = outputs["instances"].to("cpu")._fields["scores"].tolist()
        for bbox_idx in range(len(boxes)):
            if labels[bbox_idx] in ignore_catids:
                continue
            page_layout_result["layout_dets"].append({
                "category_id": labels[bbox_idx],
                "poly": [
                    boxes[bbox_idx][0], boxes[bbox_idx][1],
                    boxes[bbox_idx][2], boxes[bbox_idx][1],
                    boxes[bbox_idx][2], boxes[bbox_idx][3],
                    boxes[bbox_idx][0], boxes[bbox_idx][3],
                ],
                "score": scores[bbox_idx]
            })
        return page_layout_result


import json
import os


def read_config():
    home_dir = os.path.expanduser("~")

    config_file = os.path.join(home_dir, "magic-doc.json")

    if not os.path.exists(config_file):
        raise Exception(f"{config_file} not found")

    with open(config_file, "r") as f:
        config = json.load(f)
    return config


def get_s3_config(bucket_name: str):
    """
    从 ~/magic-doc.json 读取配置
    """
    config = read_config()

    bucket_info = config.get("bucket_info")
    if bucket_name not in bucket_info:
        access_key, secret_key, storage_endpoint = bucket_info["[default]"]
    else:
        access_key, secret_key, storage_endpoint = bucket_info[bucket_name]

    if access_key is None or secret_key is None or storage_endpoint is None:
        raise Exception("ak, sk or endpoint not found in magic-doc.json")

    # logger.info(f"get_s3_config: ak={access_key}, sk={secret_key}, endpoint={storage_endpoint}")

    return access_key, secret_key, storage_endpoint




from magic_doc.model.parallel_paddle import ParallelPaddle


class PaddleDocAnalysis:
    def __init__(self, **kwargs):
        self.model = ParallelPaddle(**kwargs)

    def __call__(self, image_dicts):
        images = [(i, image_dicts[i]["img"]) for i in range(len(image_dicts))]
        results = sorted(self.model(images), key=lambda x: x[0])
        if len(results) != len(image_dicts):
            raise Exception("fatal error: failed to inference using paddleocr")

        model_json = []
        for index, img_dict in enumerate(image_dicts):
            img = img_dict["img"]
            page_width = img_dict["width"]
            page_height = img_dict["height"]
            page_info = {"page_no": index, "height": page_height, "width": page_width}
            page_dict = {"layout_dets": results[index][1], "page_info": page_info}
            model_json.append(page_dict)
        return model_json




import json
import sys

from pathlib import Path
from typing import List

from loguru import logger
from pptx import Presentation
from pptx.enum.shapes import MSO_SHAPE_TYPE
from pptx.presentation import Presentation as ppt
from pptx.shapes.autoshape import Shape
from pptx.shapes.graphfrm import GraphicFrame
from pptx.table import Table, _Row, _Cell
from pptx.slide import Slide
from pptx.shapes.group import GroupShape
from werkzeug.datastructures import FileStorage

from magic_doc.contrib.office import OfficeExtractor
from magic_doc.contrib.model import ExtractResponse, Page, Content


class PptxExtractor(OfficeExtractor):
    def __init__(self) -> None:
        super().__init__()

    def setup(self):
        pass

    def handle_shape(
        self,
        shape: Shape,
        content_list: List[Content],
    ):
        if shape.has_text_frame:
            for paragraph in shape.text_frame.paragraphs:
                content_list.append(
                    Content(
                        type="text",
                        data=paragraph.text + "\n",
                    )
                )
        elif shape.shape_type == MSO_SHAPE_TYPE.PICTURE:
            pass
        elif shape.shape_type == MSO_SHAPE_TYPE.TABLE:
            shape: GraphicFrame
            table: Table = shape.table
            md = "\n"
            for row_no, row in enumerate(table.rows):
                row: _Row
                md += "|"
                if row_no == 1:
                    for col in row.cells:
                        md += "---|"
                    md += "\n|"
                for col in row.cells:
                    cell: _Cell = col
                    md += " " + cell.text.replace("\r", " ").replace("\n", " ") + " |"
                md += "\n"
            md += "\n"
            content_list.append(Content(type="md", data=md))
        elif shape.shape_type == MSO_SHAPE_TYPE.GROUP:
            shape: GroupShape
            for sub_shape in shape.shapes:
                self.handle_shape(sub_shape, content_list)
        else:
            # print(shape.shape_type, type(shape), file=sys.stderr)
            pass

    def extract(
        self,
        r: FileStorage | Path,
    ) -> ExtractResponse:
        pages = []
   
        presentation: ppt = Presentation(r)
        for page_no, slide in enumerate(presentation.slides):
            slide: Slide
            page = Page(page_no=page_no, content_list=[])
            for shape in slide.shapes:
                self.handle_shape(
                    shape,
                    page["content_list"],
                )

            pages.append(page)
        return pages




'''
Author: dbliu shaxunyeman@gmail.com
Date: 2024-09-02 17:55:19
LastEditors: dbliu shaxunyeman@gmail.com
LastEditTime: 2024-09-04 09:37:59
FilePath: /magic-doc/test/test_docconv.py
Description: 
'''

from magic_doc.docconv import ConvException, S3Config, DocConverter

def test_conv_localfile():
    convert = DocConverter(None, models_dir='/Users/dbliu/work/machine-learn/PDF-Extract-Kit/models')
    many_docs = ["/path/docs/mypdf.pdf", "/path/docs/mydoc.docx", "/path/docs/mydoc.doc", "/path/docs/mydoc.pptx", "/path/docs/mydoc.ppt"]
    for i, doc in enumerate(many_docs):
        try:
            markdown = convert.convert(doc, f"/path/progress/progress-{i}.txt")
            # do something with markdown
        except ConvException as e:
            assert False, f"Failed to convert {doc}, Reason: {e.message}"
        except Exception as e:
            assert False, f"Failed to convert {doc}: {e}"
        
        
def test_conv_s3file():
    s3cfg = S3Config("ak", "sk", "endpoint")
    convert = DocConverter(s3cfg, models_dir='/Users/dbliu/work/machine-learn/PDF-Extract-Kit/models')
    many_docs = ["s3://bucket/mypdf.pdf", "s3://bucket/mydoc.docx", "s3://bucket/mydoc.doc", "s3://bucket/mydoc.pptx", "s3://bucket/mydoc.ppt"]
    for i, doc in enumerate(many_docs):
        try:
            markdown = convert.convert(doc, f"/path/progress/progress-{i}.txt")
            # do something with markdown
        except ConvException as e:
            assert False, f"Failed to convert {doc}, Reason: {e.message}"
        except Exception as e:
            assert False, f"Failed to convert {doc}: {e}"
            


'''
Author: dbliu shaxunyeman@gmail.com
Date: 2024-09-02 17:55:19
LastEditors: dbliu shaxunyeman@gmail.com
LastEditTime: 2024-09-05 09:00:49
FilePath: /magic-doc/magic_doc/conv/base.py
Description: 
'''

from abc import ABC, abstractmethod

from magic_doc.progress.pupdator import ConvProgressUpdator
from magic_pdf.rw.AbsReaderWriter import AbsReaderWriter
class BaseConv(ABC):
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    def to_md(self, bits: bytes | str, pupdator:ConvProgressUpdator, **kwargs) -> str:
        return NotImplemented

    def to_mid_result(self, rw: AbsReaderWriter,  bits: bytes | str, pupdator:ConvProgressUpdator) -> list[dict] | dict:
        pupdator.update(100)
        return {}


class ParseFailed(Exception):
    pass  


'''
Author: dbliu shaxunyeman@gmail.com
Date: 2024-09-02 17:55:19
LastEditors: dbliu shaxunyeman@gmail.com
LastEditTime: 2024-09-05 10:03:21
FilePath: /magic-doc/magic_doc/common/default_config.py
Description: 
'''

from magic_doc.utils.yaml_load import patch_dict_with_env

class PdfFastParseMethod:
    AUTO = "auto"
    FAST = "fast"
    LITEOCR = "lite_ocr"

class PdfHqParseMethod:
    AUTO = "auto"
    OCR = "ocr"
    TXT = "txt"


DEFAULT_CONFIG = {
    "pdf": {
        "fast": {
            "parsemethod": PdfFastParseMethod.AUTO,
            "liteocrmodelinstance": 1,
        }, 
        "hq": {
            "parsemethod": PdfHqParseMethod.AUTO,
        }
    }
}


DEFAULT_CONFIG = patch_dict_with_env("filter", DEFAULT_CONFIG)




import shutil

from abc import ABC, abstractmethod

from concurrent.futures import Future
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from pathlib import Path
from typing import Tuple

from loguru import logger
from werkzeug.datastructures import FileStorage

# from pedia_document_parser.config import Config
from magic_doc.contrib.model import ExtractResponse, Extractor
# from pedia_document_parser.s3.client import S3Client


class OfficeExtractor(Extractor, ABC):
    def __init__(self) -> None:
        super().__init__()
        # self.config = Config()
        self.tpe = ThreadPoolExecutor(max_workers=30)
        self.counter = {}
        self.tmp_dir = Path("/tmp")
        self.max_text_count = 50_0000

    # def generate_img_path(self, id: str, image_name: str) -> str:
    #     return f"s3://{self.config.s3_bucket}/{datetime.today().strftime('%Y-%m-%d')}/{id}/{image_name}"
    #
    # def upload(self, id: str, s3_path: str, path: Path) -> Tuple[str, str]:
    #     cli = S3Client(self.config.s3_ak, self.config.s3_sk, self.config.s3_ep)
    #     cli.upload_file(s3_path, path.absolute().as_posix())
    #     return (id, s3_path)

    # def upload_background(self, id: str, img_map: dict[Path, str]):
    #     if len(img_map) == 0:
    #         self.clean_up(id)
    #         return
    #
    #     self.counter[id] = len(img_map)
    #     for src, dest in img_map.items():
    #         fut = self.tpe.submit(self.upload, id, dest, src)
    #         fut.add_done_callback(self.on_upload_succ)

    def clean_up(self, id: str):
        dir = self.get_dir_by_id(id).absolute().as_posix()
        shutil.rmtree(dir)
        self.counter.pop(id, 0)
        logger.debug(f"del {dir}")

    def on_upload_succ(self, fut: Future[Tuple[str, str]]) -> None:
        id, s3_path = fut.result()
        logger.debug(f"upload {s3_path} succ")

        self.counter[id] -= 1
        if self.counter[id] == 0:
            self.clean_up(id)

    def wait_all(self):
        self.tpe.shutdown(wait=True)

    def get_dir_by_id(self, id: str) -> Path:
        return self.tmp_dir.joinpath(id)

    def run(self, id: str, r: FileStorage, skip_image: bool = True) -> ExtractResponse:
        dir = self.get_dir_by_id(id)

        dir.mkdir()
        media_dir = dir.joinpath("media")
        media_dir.mkdir()

        try:
            return self.extract(r, id, dir, media_dir, skip_image)
        except Exception as e:
            self.clean_up(id)
            raise e

    @abstractmethod
    def extract(
        self,
        r: FileStorage | Path,
        id: str,
        dir: Path,
        media_dir: Path,
        skip_image: bool,
    ) -> ExtractResponse:
        pass




from magic_pdf.model.pp_structure_v2 import CustomPaddleModel
from magic_doc.utils import split_to_chunks
import paddle 
from concurrent.futures import ThreadPoolExecutor, as_completed


class SeqPaddle:
    def __init__(self, **kwargs):
        self.model = CustomPaddleModel(ocr=True, show_log=False)

    def __call__(self, params):
        """
        params: list[(idx, image, *args)]
        """
        results = [] 
        for idx, img in params:
            ocr_res = self.model(img) or []
            results.append((idx, ocr_res))

        return results 



import os

from magic_pdf.libs.MakeContentConfig import DropMode, MakeMode
from magic_pdf.pipe.UNIPipe import UNIPipe
from magic_pdf.pipe.OCRPipe import OCRPipe
from magic_pdf.pipe.TXTPipe import TXTPipe
from magic_doc.conv.base import BaseConv

from magic_doc.progress.filepupdator import FileBaseProgressUpdator
from magic_doc.progress.pupdator import ConvProgressUpdator
from magic_doc.utils import get_repo_directory
from magic_doc.utils.null_writer import NullWriter
from magic_pdf.dict2md.ocr_mkcontent import union_make
from magic_pdf.libs.json_compressor import JsonCompressor
from magic_pdf.rw.AbsReaderWriter import AbsReaderWriter
from magic_pdf.rw.DiskReaderWriter import DiskReaderWriter
from magic_pdf.tools.common import prepare_env
from magic_doc.common.default_config import DEFAULT_CONFIG, PdfHqParseMethod


NULL_IMG_DIR = "/tmp"

class SingletonModelWrapper:

    def __new__(cls, models_dir):
        if not hasattr(cls, "instance"):
            from magic_doc.model.doc_analysis import DocAnalysis
            apply_ocr = os.getenv("APPLY_OCR", "TRUE") == "TRUE" 
            apply_layout = os.getenv("APPLY_LAYOUT", "TRUE") == "TRUE" 
            apply_formula = os.getenv("APPLY_FORMULA", "FALSE") == "TRUE"
            apply_table = os.getenv("APPLY_TABLE", "FALSE") == "TRUE"
            
            cls.instance = super(SingletonModelWrapper, cls).__new__(cls)
            cls.instance.doc_analysis = DocAnalysis(  # type: ignore
                models_dir=models_dir,
                configs=os.path.join(
                    get_repo_directory(), "resources/model/model_configs.yaml"
                ),
                apply_ocr=apply_ocr, 
                apply_layout=apply_layout, 
                apply_formula=apply_formula,
                table_config={
                    'is_table_recog_enable': apply_table,
                    'max_time': 400,
                }
            )
        return cls.instance
    
    def __call__(self, bits: bytes):
        from magic_pdf.model.doc_analyze_by_custom_model import load_images_from_pdf
        images = load_images_from_pdf(bits, dpi=200)
        return self.doc_analysis(images) # type: ignore

class Pdf(BaseConv):

    def __init__(self, models_dir):
        super().__init__()
        self.models_dir = models_dir

    def __construct_pdf_pipe(self, bits, model_list, image_writer, parse_method):
        if parse_method == PdfHqParseMethod.AUTO:
            jso_useful_key = {"_pdf_type": "", "model_list": model_list}
            pipe = UNIPipe(bits, jso_useful_key, image_writer, is_debug=True)  # type: ignore
        elif parse_method == PdfHqParseMethod.OCR:
            pipe = OCRPipe(bits, model_list, image_writer, is_debug=True)  # type: ignore
        elif parse_method == PdfHqParseMethod.TXT:
            pipe = TXTPipe(bits, model_list, image_writer, is_debug=True)  # type: ignore
        else:
            raise Exception("unknown parse method under hq mode")
        return pipe


    def to_md(self, bits: bytes | str, pupdator: ConvProgressUpdator, **kwargs) -> str:

        output_dir = kwargs.get('output', None)
        pdf_file_name = kwargs.get('file_name', None)
        parse_method = kwargs.get('method', None)
        parse_method = parse_method if parse_method else DEFAULT_CONFIG["pdf"]["hq"]["parsemethod"]

        if output_dir and pdf_file_name:
            local_image_dir, local_md_dir = prepare_env(
                output_dir, 
                pdf_file_name, 
                parse_method,
            )
            image_writer, md_writer = DiskReaderWriter(local_image_dir), DiskReaderWriter(local_md_dir)
            image_dir = str(os.path.basename(local_image_dir))
        else:
            image_writer = NullWriter()
            md_writer = None
            image_dir = None

        model_proc = SingletonModelWrapper(self.models_dir)
        pupdator.update(0)

        model_list = model_proc(bits)  # type: ignore
        pupdator.update(50)
        
        pipe = self.__construct_pdf_pipe(bits, model_list, image_writer, parse_method)
        pipe.pipe_classify() # 默认ocrpipe的时候不需要再做分类，可以节省时间
        pipe.pipe_parse()
        pupdator.update(100)

        if not image_dir:
            pdf_mid_data = JsonCompressor.decompress_json(pipe.get_compress_pdf_mid_data())
            pdf_info_list = pdf_mid_data["pdf_info"]
            md_content = union_make(pdf_info_list, MakeMode.NLP_MD, DropMode.NONE, NULL_IMG_DIR)
        else:
            md_content = pipe.pipe_mk_markdown(
                image_dir, 
                drop_mode=DropMode.NONE, 
                md_make_mode=MakeMode.MM_MD
            )
            if md_writer:
                md_writer.write(
                    content=md_content,
                    path=f"{pdf_file_name}.md",
                    mode=AbsReaderWriter.MODE_TXT,
                )

        return md_content # type: ignore

    def to_mid_result(self, image_writer: AbsReaderWriter, bits: bytes | str, pupdator: ConvProgressUpdator) -> list[dict] | dict:
        model_proc = SingletonModelWrapper()
        pupdator.update(0)

        model_list = model_proc(bits)  # type: ignore
        pupdator.update(50)
        # jso_useful_key = {
        #     "_pdf_type": "",
        #     "model_list": model_list,
        # }
        pipe = self.__construct_pdf_pipe(bits, model_list, image_writer)
        # pipe.pipe_classify()
        pipe.pipe_parse()
        pupdator.update(100)

        pdf_mid_data = JsonCompressor.decompress_json(pipe.get_compress_pdf_mid_data())
        pdf_info_list = pdf_mid_data["pdf_info"]
        return pdf_info_list

if __name__ == "__main__":
    with open("/opt/data/pdf/20240423/pdf_test2/ol006018w.pdf", "rb") as f:
        bits_data = f.read()
        parser = Pdf(models_dir='/Users/dbliu/work/machine-learn/PDF-Extract-Kit/models')
        md_content = parser.to_md(
            bits_data, FileBaseProgressUpdator("debug/progress.txt")
        )
        with open("debug/pdf2md.by_model.md", "w") as f:
            f.write(md_content) # type: ignore




import torch
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Union

from transformers import BatchEncoding, PreTrainedTokenizerBase
from transformers.data.data_collator import (
    DataCollatorMixin,
    _torch_collate_batch,
)
from transformers.file_utils import PaddingStrategy

from typing import NewType
InputDataClass = NewType("InputDataClass", Any)

def pre_calc_rel_mat(segment_ids):
    valid_span = torch.zeros((segment_ids.shape[0], segment_ids.shape[1], segment_ids.shape[1]),
                             device=segment_ids.device, dtype=torch.bool)
    for i in range(segment_ids.shape[0]):
        for j in range(segment_ids.shape[1]):
            valid_span[i, j, :] = segment_ids[i, :] == segment_ids[i, j]

    return valid_span

@dataclass
class DataCollatorForKeyValueExtraction(DataCollatorMixin):
    """
    Data collator that will dynamically pad the inputs received, as well as the labels.
    Args:
        tokenizer (:class:`~transformers.PreTrainedTokenizer` or :class:`~transformers.PreTrainedTokenizerFast`):
            The tokenizer used for encoding the data.
        padding (:obj:`bool`, :obj:`str` or :class:`~transformers.file_utils.PaddingStrategy`, `optional`, defaults to :obj:`True`):
            Select a strategy to pad the returned sequences (according to the model's padding side and padding index)
            among:
            * :obj:`True` or :obj:`'longest'`: Pad to the longest sequence in the batch (or no padding if only a single
              sequence if provided).
            * :obj:`'max_length'`: Pad to a maximum length specified with the argument :obj:`max_length` or to the
              maximum acceptable input length for the model if that argument is not provided.
            * :obj:`False` or :obj:`'do_not_pad'` (default): No padding (i.e., can output a batch with sequences of
              different lengths).
        max_length (:obj:`int`, `optional`):
            Maximum length of the returned list and optionally padding length (see above).
        pad_to_multiple_of (:obj:`int`, `optional`):
            If set will pad the sequence to a multiple of the provided value.
            This is especially useful to enable the use of Tensor Cores on NVIDIA hardware with compute capability >=
            7.5 (Volta).
        label_pad_token_id (:obj:`int`, `optional`, defaults to -100):
            The id to use when padding the labels (-100 will be automatically ignore by PyTorch loss functions).
    """

    tokenizer: PreTrainedTokenizerBase
    padding: Union[bool, str, PaddingStrategy] = True
    max_length: Optional[int] = None
    pad_to_multiple_of: Optional[int] = None
    label_pad_token_id: int = -100

    def __call__(self, features):
        label_name = "label" if "label" in features[0].keys() else "labels"
        labels = [feature[label_name] for feature in features] if label_name in features[0].keys() else None

        images = None
        if "images" in features[0]:
            images = torch.stack([torch.tensor(d.pop("images")) for d in features])
            IMAGE_LEN = int(images.shape[-1] / 16) * int(images.shape[-1] / 16) + 1

        batch = self.tokenizer.pad(
            features,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            # Conversion to tensors will fail if we have labels as they are not of the same length yet.
            return_tensors="pt" if labels is None else None,
        )

        if images is not None:
            batch["images"] = images
            batch = {k: torch.tensor(v, dtype=torch.int64) if isinstance(v[0], list) and k == 'attention_mask' else v
                     for k, v in batch.items()}
            visual_attention_mask = torch.ones((len(batch['input_ids']), IMAGE_LEN), dtype=torch.long)
            batch["attention_mask"] = torch.cat([batch['attention_mask'], visual_attention_mask], dim=1)

        if labels is None:
            return batch

        has_bbox_input = "bbox" in features[0]
        has_position_input = "position_ids" in features[0]
        padding_idx=self.tokenizer.pad_token_id
        sequence_length = torch.tensor(batch["input_ids"]).shape[1]
        padding_side = self.tokenizer.padding_side
        if padding_side == "right":
            batch["labels"] = [label + [self.label_pad_token_id] * (sequence_length - len(label)) for label in labels]
            if has_bbox_input:
                batch["bbox"] = [bbox + [[0, 0, 0, 0]] * (sequence_length - len(bbox)) for bbox in batch["bbox"]]
            if has_position_input:
                batch["position_ids"] = [position_id + [padding_idx] * (sequence_length - len(position_id))
                                          for position_id in batch["position_ids"]]

        else:
            batch["labels"] = [[self.label_pad_token_id] * (sequence_length - len(label)) + label for label in labels]
            if has_bbox_input:
                batch["bbox"] = [[[0, 0, 0, 0]] * (sequence_length - len(bbox)) + bbox for bbox in batch["bbox"]]
            if has_position_input:
                batch["position_ids"] = [[padding_idx] * (sequence_length - len(position_id))
                                          + position_id for position_id in batch["position_ids"]]

        if 'segment_ids' in batch:
            assert 'position_ids' in batch
            for i in range(len(batch['segment_ids'])):
                batch['segment_ids'][i] = batch['segment_ids'][i] + [batch['segment_ids'][i][-1] + 1] * (sequence_length - len(batch['segment_ids'][i])) + [
                    batch['segment_ids'][i][-1] + 2] * IMAGE_LEN

        batch = {k: torch.tensor(v, dtype=torch.int64) if isinstance(v[0], list) else v for k, v in batch.items()}

        if 'segment_ids' in batch:
            valid_span = pre_calc_rel_mat(
                segment_ids=batch['segment_ids']
            )
            batch['valid_span'] = valid_span
            del batch['segment_ids']

        if images is not None:
            visual_labels = torch.ones((len(batch['input_ids']), IMAGE_LEN), dtype=torch.long) * -100
            batch["labels"] = torch.cat([batch['labels'], visual_labels], dim=1)

        return batch



import xml.etree.ElementTree as ET
import zipfile

from pathlib import Path
from magic_doc.contrib.model import ExtractResponse, Content, Page
from magic_doc.contrib.office import OfficeExtractor
from typing import IO
from io import BytesIO
from werkzeug.datastructures import FileStorage

from magic_doc.contrib.office.formula.omml import omml2tex


class DocxExtractor(OfficeExtractor):
    def __init__(self) -> None:
        super().__init__()

    def setup(self):
        pass

    def __word2markdown(
        self,
        docx_file_stream: IO[bytes],
    ):
        tag_w = "{http://schemas.openxmlformats.org/wordprocessingml/2006/main}"
        tag_body = f"{tag_w}body"

        content_list = []
        text_count = 0
        with zipfile.ZipFile(docx_file_stream, "r") as docx:
            xml_content = docx.read("word/document.xml")
            tree = ET.XML(xml_content)
            body = tree.find(tag_body)

            for child in body:
                tag = child.tag.split("}")[-1]
                if text_count >= self.max_text_count:
                    break

                match tag:
                    case "p":
                        text = ""
                        for ele in child.iter():
                            if "math" in ele.tag:
                                if ele.tag.endswith("oMath"):
                                    math_xml = BytesIO()
                                    ET.ElementTree(ele).write(
                                        math_xml,
                                        encoding="utf-8",
                                        xml_declaration=True,
                                    )
                                    math_xml = math_xml.getvalue().decode("utf-8")
                                    math_xml = "\n".join(math_xml.split("\n")[1:])
                                    math_formula = "\n" + omml2tex(math_xml) + "\n"

                                    text = text.strip()
                                    if len(text) > 0:
                                        text_count += len(text) + 1
                                        content_list.append(
                                            Content(type="text", data=text)
                                        )
                                        text = ""

                                    text_count += len(math_formula)
                                    content_list.append(
                                        Content(type="md", data=math_formula)
                                    )
                                continue
                            if t := ele.text:
                                if len(t) > 0:
                                    text += t
                        text = text.strip()
                        if len(text) > 0:
                            text_count += len(text) + 1
                            content_list.append(Content(type="text", data=text))
                            text = ""
                    case "tbl":
                        col_size = len(list(child.find(f"{tag_w}tblGrid")))
                        md = "\n"
                        for idx, row in enumerate(child.iter(f"{tag_w}tr")):
                            if idx == 1:
                                md += "|"
                                for _ in range(col_size):
                                    md += "---|"
                                md += "\n"
                            md += "|"
                            # print(row)
                            for cell in row.iter(f"{tag_w}tc"):
                                t = ""
                                for cell_ele in cell.itertext():
                                    t += (
                                        cell_ele.strip()
                                        .replace("\r", "")
                                        .replace("\n", "")
                                    )
                                md += f" {t} |"
                            md += "\n"
                        md += "\n"
                        text_count += len(md)
                        content_list.append(Content(type="md", data=md))
                    case "sectPr":
                        # docx section pointer, meaningless
                        pass
                    case unknown:
                        pass
                        # print(unknown)
            return content_list

    def extract(
        self,
        r: FileStorage | Path,
        id: str,
        dir: Path,
        media_dir: Path,
        skip_image: bool,
    ) -> ExtractResponse:
        if type(r) is FileStorage:
            page = Page(
                page_no=0,
                content_list=self.__word2markdown(r.stream),
            )
        else:
            page = Page(
                page_no=0,
                content_list=self.__word2markdown(open(r, "rb")),
            )
        return [page]


if __name__ == "__main__":
    e = DocxExtractor()

    res = e.run(
        "def",
        Path(
            "test_data/doc/【中简】模电自测第四版.docx",
        ),
    )

    print(res)

    e.wait_all()



from magic_pdf.libs.pdf_check import detect_invalid_chars

import magic_doc
import os
import random
import fitz

def get_repo_directory():
    return os.path.dirname(magic_doc.__file__)


def is_digital(bits: bytes) -> bool:
    def _is_digital(doc, check_page=10, text_len_thrs=100) -> bool:
        sample_page_num = min(check_page, doc.page_count)
        page_ids = random.sample(range(doc.page_count), sample_page_num)
        page_text_len = [
            len(doc[pno].get_text("text")) > text_len_thrs for pno in page_ids
        ]
        if any(page_text_len):
            return True
        return False

    def _check_invalid_chars(pdf_bytes: bytes) -> bool:
        return detect_invalid_chars(pdf_bytes)
    
    with fitz.open(stream=bits) as doc:
        return _is_digital(doc) and _check_invalid_chars(bits)


def split_to_chunks(lst, n):
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


# coding=utf-8
from transformers.models.bert.configuration_bert import BertConfig
from transformers.utils import logging


logger = logging.get_logger(__name__)

LAYOUTLMV3_PRETRAINED_CONFIG_ARCHIVE_MAP = {
    "layoutlmv3-base": "https://huggingface.co/microsoft/layoutlmv3-base/resolve/main/config.json",
    "layoutlmv3-large": "https://huggingface.co/microsoft/layoutlmv3-large/resolve/main/config.json",
    # See all LayoutLMv3 models at https://huggingface.co/models?filter=layoutlmv3
}


class LayoutLMv3Config(BertConfig):
    model_type = "layoutlmv3"

    def __init__(
        self,
        pad_token_id=1,
        bos_token_id=0,
        eos_token_id=2,
        max_2d_position_embeddings=1024,
        coordinate_size=None,
        shape_size=None,
        has_relative_attention_bias=False,
        rel_pos_bins=32,
        max_rel_pos=128,
        has_spatial_attention_bias=False,
        rel_2d_pos_bins=64,
        max_rel_2d_pos=256,
        visual_embed=True,
        mim=False,
        wpa_task=False,
        discrete_vae_weight_path='',
        discrete_vae_type='dall-e',
        input_size=224,
        second_input_size=112,
        device='cuda',
        **kwargs
    ):
        """Constructs RobertaConfig."""
        super().__init__(pad_token_id=pad_token_id, bos_token_id=bos_token_id, eos_token_id=eos_token_id, **kwargs)
        self.max_2d_position_embeddings = max_2d_position_embeddings
        self.coordinate_size = coordinate_size
        self.shape_size = shape_size
        self.has_relative_attention_bias = has_relative_attention_bias
        self.rel_pos_bins = rel_pos_bins
        self.max_rel_pos = max_rel_pos
        self.has_spatial_attention_bias = has_spatial_attention_bias
        self.rel_2d_pos_bins = rel_2d_pos_bins
        self.max_rel_2d_pos = max_rel_2d_pos
        self.visual_embed = visual_embed
        self.mim = mim
        self.wpa_task = wpa_task
        self.discrete_vae_weight_path = discrete_vae_weight_path
        self.discrete_vae_type = discrete_vae_type
        self.input_size = input_size
        self.second_input_size = second_input_size
        self.device = device



# Copyright (c) Facebook, Inc. and its affiliates.
import logging
import numpy as np
from typing import Dict, List, Optional, Tuple
import torch
from torch import nn

from detectron2.config import configurable
from detectron2.structures import ImageList, Instances
from detectron2.utils.events import get_event_storage

from detectron2.modeling.backbone import Backbone, build_backbone
from detectron2.modeling.meta_arch.build import META_ARCH_REGISTRY

from detectron2.modeling.meta_arch import GeneralizedRCNN

from detectron2.modeling.postprocessing import detector_postprocess
from detectron2.modeling.roi_heads.fast_rcnn import fast_rcnn_inference_single_image
from contextlib import contextmanager
from itertools import count

@META_ARCH_REGISTRY.register()
class VLGeneralizedRCNN(GeneralizedRCNN):
    """
    Generalized R-CNN. Any models that contains the following three components:
    1. Per-image feature extraction (aka backbone)
    2. Region proposal generation
    3. Per-region feature extraction and prediction
    """

    def forward(self, batched_inputs: List[Dict[str, torch.Tensor]]):
        """
        Args:
            batched_inputs: a list, batched outputs of :class:`DatasetMapper` .
                Each item in the list contains the inputs for one image.
                For now, each item in the list is a dict that contains:

                * image: Tensor, image in (C, H, W) format.
                * instances (optional): groundtruth :class:`Instances`
                * proposals (optional): :class:`Instances`, precomputed proposals.

                Other information that's included in the original dicts, such as:

                * "height", "width" (int): the output resolution of the model, used in inference.
                  See :meth:`postprocess` for details.

        Returns:
            list[dict]:
                Each dict is the output for one input image.
                The dict contains one key "instances" whose value is a :class:`Instances`.
                The :class:`Instances` object has the following keys:
                "pred_boxes", "pred_classes", "scores", "pred_masks", "pred_keypoints"
        """
        if not self.training:
            return self.inference(batched_inputs)

        images = self.preprocess_image(batched_inputs)
        if "instances" in batched_inputs[0]:
            gt_instances = [x["instances"].to(self.device) for x in batched_inputs]
        else:
            gt_instances = None

        # features = self.backbone(images.tensor)
        input = self.get_batch(batched_inputs, images)
        features = self.backbone(input)

        if self.proposal_generator is not None:
            proposals, proposal_losses = self.proposal_generator(images, features, gt_instances)
        else:
            assert "proposals" in batched_inputs[0]
            proposals = [x["proposals"].to(self.device) for x in batched_inputs]
            proposal_losses = {}

        _, detector_losses = self.roi_heads(images, features, proposals, gt_instances)
        if self.vis_period > 0:
            storage = get_event_storage()
            if storage.iter % self.vis_period == 0:
                self.visualize_training(batched_inputs, proposals)

        losses = {}
        losses.update(detector_losses)
        losses.update(proposal_losses)
        return losses

    def inference(
        self,
        batched_inputs: List[Dict[str, torch.Tensor]],
        detected_instances: Optional[List[Instances]] = None,
        do_postprocess: bool = True,
    ):
        """
        Run inference on the given inputs.

        Args:
            batched_inputs (list[dict]): same as in :meth:`forward`
            detected_instances (None or list[Instances]): if not None, it
                contains an `Instances` object per image. The `Instances`
                object contains "pred_boxes" and "pred_classes" which are
                known boxes in the image.
                The inference will then skip the detection of bounding boxes,
                and only predict other per-ROI outputs.
            do_postprocess (bool): whether to apply post-processing on the outputs.

        Returns:
            When do_postprocess=True, same as in :meth:`forward`.
            Otherwise, a list[Instances] containing raw network outputs.
        """
        assert not self.training

        images = self.preprocess_image(batched_inputs)
        # features = self.backbone(images.tensor)
        input = self.get_batch(batched_inputs, images)
        features = self.backbone(input)

        if detected_instances is None:
            if self.proposal_generator is not None:
                proposals, _ = self.proposal_generator(images, features, None)
            else:
                assert "proposals" in batched_inputs[0]
                proposals = [x["proposals"].to(self.device) for x in batched_inputs]

            results, _ = self.roi_heads(images, features, proposals, None)
        else:
            detected_instances = [x.to(self.device) for x in detected_instances]
            results = self.roi_heads.forward_with_given_boxes(features, detected_instances)

        if do_postprocess:
            assert not torch.jit.is_scripting(), "Scripting is not supported for postprocess."
            return GeneralizedRCNN._postprocess(results, batched_inputs, images.image_sizes)
        else:
            return results

    def get_batch(self, examples, images):
        if len(examples) >= 1 and "bbox" not in examples[0]:  # image_only
            return {"images": images.tensor}

        return input

    def _batch_inference(self, batched_inputs, detected_instances=None):
        """
        Execute inference on a list of inputs,
        using batch size = self.batch_size (e.g., 2), instead of the length of the list.

        Inputs & outputs have the same format as :meth:`GeneralizedRCNN.inference`
        """
        if detected_instances is None:
            detected_instances = [None] * len(batched_inputs)

        outputs = []
        inputs, instances = [], []
        for idx, input, instance in zip(count(), batched_inputs, detected_instances):
            inputs.append(input)
            instances.append(instance)
            if len(inputs) == 2 or idx == len(batched_inputs) - 1:
                outputs.extend(
                    self.inference(
                        inputs,
                        instances if instances[0] is not None else None,
                        do_postprocess=True,  # False
                    )
                )
                inputs, instances = [], []
        return outputs



from transformers import AutoConfig, AutoModel, AutoModelForTokenClassification, \
    AutoModelForQuestionAnswering, AutoModelForSequenceClassification, AutoTokenizer
from transformers.convert_slow_tokenizer import SLOW_TO_FAST_CONVERTERS, RobertaConverter

from .configuration_layoutlmv3 import LayoutLMv3Config
from .modeling_layoutlmv3 import (
    LayoutLMv3ForTokenClassification,
    LayoutLMv3ForQuestionAnswering,
    LayoutLMv3ForSequenceClassification,
    LayoutLMv3Model,
)
from .tokenization_layoutlmv3 import LayoutLMv3Tokenizer
from .tokenization_layoutlmv3_fast import LayoutLMv3TokenizerFast


#AutoConfig.register("layoutlmv3", LayoutLMv3Config)
#AutoModel.register(LayoutLMv3Config, LayoutLMv3Model)
#AutoModelForTokenClassification.register(LayoutLMv3Config, LayoutLMv3ForTokenClassification)
#AutoModelForQuestionAnswering.register(LayoutLMv3Config, LayoutLMv3ForQuestionAnswering)
#AutoModelForSequenceClassification.register(LayoutLMv3Config, LayoutLMv3ForSequenceClassification)
#AutoTokenizer.register(
#    LayoutLMv3Config, slow_tokenizer_class=LayoutLMv3Tokenizer, fast_tokenizer_class=LayoutLMv3TokenizerFast
#)
SLOW_TO_FAST_CONVERTERS.update({"LayoutLMv3Tokenizer": RobertaConverter})



from magic_pdf.model.pp_structure_v2 import CustomPaddleModel
from magic_doc.utils import split_to_chunks
import paddle 
from concurrent.futures import ThreadPoolExecutor, as_completed
import math

class ParallelPaddle:
    def __init__(self, model_load_on_each_gpu_count=1):
        models = []
        for _ in range(model_load_on_each_gpu_count):
            models.append(CustomPaddleModel(ocr=True, show_log=False))
        self.models = models

    def __call__(self, params):
        """
        params: list[(idx, image, *args)]
        """
        if len(params) == 0:
            return []
        chunks = list(split_to_chunks(params, max(math.ceil(len(params) *1.0/ len(self.models)), 1)))
        return self._run_ocr_concurrently(chunks)


    def _run_ocr_concurrently(self, chunks):
        results = []
        def run_ocr(chunk, i):
            result = []
            for idx, img in chunk:
                ocr_res = self.models[i](img) or []
                result.append((idx, ocr_res))
            return result

        with ThreadPoolExecutor(max_workers=len(chunks)) as executor:
            future_to_ocr = {executor.submit(run_ocr, chunk, i): i for i, chunk in enumerate(chunks)}
            for future in as_completed(future_to_ocr):
                try:
                    data = future.result()
                    results.extend(data)
                except Exception as exc:
                    print(f"failed to process ocr, reason: ", exc)
        return sorted(results, key=lambda x: x[0])




import os
import platform
import time
from functools import partial

import boto3
from botocore.client import Config
from func_timeout import FunctionTimedOut, func_timeout
from magic_pdf.libs.path_utils import (
    remove_non_official_s3_args,
)
from magic_pdf.rw.DiskReaderWriter import DiskReaderWriter
from magic_pdf.rw.S3ReaderWriter import S3ReaderWriter
from smart_open import open

from magic_doc.conv.base import BaseConv
from magic_doc.conv.doc_antiword import Doc as Doc_antiword
from magic_doc.conv.doc_libreoffice import Doc as Doc_libreoffice
from magic_doc.conv.docx_xml_parse import Docx
from magic_doc.conv.pdf import Pdf as fastTextPdf
from magic_doc.conv.pdf_pp_structurev2 import Pdf as liteOcrPdf
from magic_doc.conv.pdf_magicpdf import Pdf as fullPdf
from magic_doc.conv.ppt_libreoffice import Ppt
from magic_doc.conv.pptx_python_pptx import Pptx
from magic_doc.progress.filepupdator import FileBaseProgressUpdator
from magic_doc.utils import is_digital
from magic_doc.conv.base import ParseFailed
from magic_doc.common.default_config import DEFAULT_CONFIG, PdfFastParseMethod
from magic_doc.contrib.wrapper_exceptions import NotSupportOcrPDFException


class ParsePDFType:
    FAST = "fast"
    FULL = "full"


class ConvException(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class S3Config(object):
    def __init__(self, ak: str, sk: str, endpoint: str, address_style: str = "path"):
        self.ak = ak
        self.sk = sk
        self.endpoint = endpoint
        self.address_style = address_style


class DocConverter(object):
    def __init__(
        self,
        s3_config: S3Config,
        temp_dir: str = "/tmp/",
        models_dir: str = '/tmp/models/',
        parse_pdf_type=ParsePDFType.FAST,
        conv_timeout=60,
    ):
        self.__s3cfg = s3_config
        if self.__s3cfg:
            self.__s3cli = boto3.client(
                service_name="s3",
                aws_access_key_id=self.__s3cfg.ak,
                aws_secret_access_key=self.__s3cfg.sk,
                endpoint_url=self.__s3cfg.endpoint,
                config=Config(
                    s3={"addressing_style": "path"}, retries={"max_attempts": 8}
                ),
            )
        self.parse_pdf_type = parse_pdf_type
        self.__temp_dir = temp_dir
        self.__conv_timeout = conv_timeout
        self.models_dir = models_dir
        self.__init_conv()

    def __init_conv(self):
        # 根据系统选择doc解析方式
        system = platform.system()
        if system == "Linux":
            self.doc_conv = Doc_antiword()
        else:
            self.doc_conv = Doc_libreoffice()
        self.docx_conv = Docx()
        self.full_pdf_conv = fullPdf(self.models_dir)

        if DEFAULT_CONFIG["pdf"]["fast"]["parsemethod"] == PdfFastParseMethod.AUTO:
            self.fast_textpdf_conv = fastTextPdf(allowed_failure=False)
        else:
            self.fast_textpdf_conv = fastTextPdf()
        self.lite_ocrpdf_conv = liteOcrPdf()
        self.ppt_conv = Ppt()
        self.pptx_conv = Pptx()

    def __select_conv(self, doc_path: str, doc_bytes: bytes):
        def check_magic_header(magic_header: bytes) -> bool:
            if doc_bytes.startswith(magic_header):
                return True
            else:
                raise ConvException("File is broken.")

        """根据文件后缀选择转换器"""
        lower_case_path = doc_path.lower()
        if lower_case_path.endswith(".doc"):
            # OLE2
            if check_magic_header(b"\xD0\xCF\x11\xE0\xA1\xB1\x1A\xE1"):
                return self.doc_conv
        elif lower_case_path.endswith(".docx"):
            # PK
            if check_magic_header(b"\x50\x4b\x03\x04"):
                return self.docx_conv
        elif lower_case_path.endswith(".pdf"):
            # %PDF
            if check_magic_header(b"%PDF"):
                if self.parse_pdf_type == ParsePDFType.FAST:
                    if DEFAULT_CONFIG["pdf"]["fast"]["parsemethod"] == PdfFastParseMethod.AUTO:
                        if is_digital(doc_bytes):
                            return self.fast_textpdf_conv
                        else:
                            return self.lite_ocrpdf_conv
                    elif DEFAULT_CONFIG["pdf"]["fast"]["parsemethod"] == PdfFastParseMethod.FAST:
                        return self.fast_textpdf_conv
                    elif DEFAULT_CONFIG["pdf"]["fast"]["parsemethod"] == PdfFastParseMethod.LITEOCR:
                        return self.lite_ocrpdf_conv
                    else:
                        return self.lite_ocrpdf_conv
                else:
                    return self.full_pdf_conv

        elif lower_case_path.endswith(".ppt"):
            # OLE2
            if check_magic_header(b"\xD0\xCF\x11\xE0\xA1\xB1\x1A\xE1"):
                return self.ppt_conv
        elif lower_case_path.endswith(".pptx"):
            # PK
            if check_magic_header(b"\x50\x4b\x03\x04"):
                return self.pptx_conv
        else:
            raise ConvException("Unsupported file format.")

    def __read_file_as_bytes(self, doc_path: str) -> bytes:
        """
        支持从本地文件和s3文件读取
        """
        if doc_path.startswith("s3://"):
            transport_params = {"client": self.__s3cli}
            with open(doc_path, "rb", transport_params=transport_params) as fin:
                content_bytes = fin.read()
                return content_bytes
        else:
            with open(doc_path, "rb") as fin:
                content_bytes = fin.read()
                return content_bytes

    def get_raw_file_content(self, doc_path:str) -> bytes:
        return self.__read_file_as_bytes(doc_path)

    def _timeout_convert(
        self,
        byte_content: bytes,
        progress_file_path: str,
        conv_method,
        conv_timeout=None,
        kwargs = None,
    ):
        """
        在线快速解析
        doc_path: str, path to the document, support local file path and s3 path.
        progress_file_path: str, path to the progress file, support local file path only.
        return markdown string or raise ConvException.
        """
        conv_timeout = conv_timeout or self.__conv_timeout  # 根据这个时间判断函数超时
        res = ""
        cost_time = 0
        try:
            prog_updator = FileBaseProgressUpdator(progress_file_path)
            start_time = time.time()
            res = func_timeout(
                conv_timeout, conv_method, args=(byte_content, prog_updator), kwargs=kwargs
            )
            end_time = time.time()
            cost_time = round(end_time - start_time, 3)

        except FunctionTimedOut as e1:
            # logger.exception(e1)
            raise ConvException("Convert timeout.")
        except ParseFailed as e2:
            raise e2
        except NotSupportOcrPDFException as e3:
            raise ParseFailed
        except Exception as ex:
            # logger.exception(e2)
            raise ConvException("Convert failed: %s" % str(ex))

        return res, cost_time

    def convert(self, doc_path: str, progress_file_path: str = "/tmp/magic_doc_convert_progress.txt", conv_timeout=None, kwargs = None):
        """
        在线快速解析
        doc_path: str, path to the document, support local file path and s3 path.
        progress_file_path: str, path to the progress file, support local file path only.
        return markdown string or raise ConvException.
        """
        byte_content = self.__read_file_as_bytes(doc_path)
        conv: BaseConv = self.__select_conv(doc_path, byte_content)
        return self._timeout_convert(
            byte_content, progress_file_path, conv.to_md, conv_timeout, kwargs
        )

    def convert_to_mid_result(
        self, doc_path: str, progress_file_path: str = "/tmp/magic_doc_convert_progress.txt", conv_timeout=None
    ):
        """
        在线快速解析
        doc_path: str, path to the document, support local file path and s3 path.
        progress_file_path: str, path to the progress file, support local file path only.
        return markdown string or raise ConvException.
        """
        byte_content = self.__read_file_as_bytes(doc_path)
        conv: BaseConv = self.__select_conv(doc_path, byte_content)

        parent_path = os.path.dirname(doc_path)
        if doc_path.startswith("s3://"):
            image_writer = S3ReaderWriter(
                self.__s3cfg.ak,
                self.__s3cfg.sk,
                self.__s3cfg.endpoint,
                "auto",
                os.path.join(
                    os.path.dirname(remove_non_official_s3_args(doc_path)), "images"
                ),
            )
        else:
            image_writer = DiskReaderWriter(os.path.join(parent_path, "images"))

        return self._timeout_convert(
            byte_content,
            progress_file_path,
            partial(conv.to_mid_result, image_writer),
            conv_timeout,
        )



import os
from datetime import datetime
import click
from pathlib import Path
from magic_doc.libs.version import __version__

from magic_doc.utils.config import get_s3_config
from magic_doc.utils.path_utils import get_local_dir, parse_s3path, prepare_env
from magic_doc.docconv import DocConverter, S3Config

from loguru import logger
from magic_doc.conv.pdf_magicpdf import SingletonModelWrapper

log_level = "ERROR"
log_dir = os.path.join(get_local_dir(), "magic-doc", "logs")
if not Path(log_dir).exists():
    Path(log_dir).mkdir(parents=True, exist_ok=True)
log_name = f'log_{datetime.now().strftime("%Y-%m-%d")}.log'
log_file_path = os.path.join(log_dir, log_name)
logger.add(str(log_file_path), rotation='00:00', encoding='utf-8', level=log_level, enqueue=True)


def abort(message=None, exit_code=1):
    click.echo(click.style(message, fg='red'))
    exit(exit_code)


total_cost_time = 0
total_convert_error = 0
total_file_broken = 0
total_unsupported_files = 0
total_time_out = 0
total_success_convert = 0

@click.group()
@click.version_option(__version__, "--version", "-v", help="显示版本信息")
def cli():
    pass

@cli.command()
# @click.version_option(__version__, "--version", "-v", help="显示版本信息")
@click.option('-f', '--file-path', 'input_file_path', type=click.STRING,
              help='file path, support s3/local/list, list file need end with ".list"')
@click.option('-p', '--progress-file-path', 'progress_file_path', default="/tmp/magic_doc_progress.txt", type=click.STRING,
              help='path to the progress file to save')
@click.option('-t', '--conv-timeout', 'conv_timeout', default=300, type=click.INT, help='timeout')
@click.option("-o", "--output", "output", default="", type=click.STRING)
@click.option("--models-dir", "models_dir", default="/tmp/models/", type=click.STRING, help="specify the models path")
def cli_conv(input_file_path, progress_file_path, conv_timeout, output, models_dir):
    global total_cost_time, total_convert_error, total_file_broken, \
        total_unsupported_files, total_time_out, total_success_convert

    def parse_doc(doc_path, pf_path=None):
        """使用全局变量统计耗时和error数量"""
        global total_cost_time, total_convert_error, total_file_broken, \
            total_unsupported_files, total_time_out, total_success_convert
        try:
            file_name = str(Path(doc_path).stem)
            '''创建同名进度缓存文件'''
            if not pf_path:
                pf_path = f"/tmp/{file_name}.txt"
            '''如果文档路径为s3链接，先获取s3配置并初始化'''
            if doc_path.startswith("s3://"):
                bucket, key = parse_s3path(doc_path)
                ak, sk, endpoint = get_s3_config(bucket)
                s3_config = S3Config(ak, sk, endpoint)
            else:
                '''非s3路径不需要初始化s3配置'''
                s3_config = None
            doc_conv = DocConverter(s3_config, models_dir=models_dir)
            markdown_string, cost_time = doc_conv.convert(doc_path, pf_path, conv_timeout)
            total_cost_time += cost_time
            logger.info(f"convert {doc_path} to markdown, cost {cost_time} seconds")
            # click.echo(markdown_string)
            base_name, doc_type = os.path.splitext(doc_path)
            out_put_dir = output or prepare_env(file_name, doc_type.lstrip(".").lower())
            with open(os.path.join(out_put_dir, file_name + ".md"), "w", encoding='utf-8') as md_file:
                md_file.write(markdown_string)
            total_success_convert += 1
            return cost_time
        except Exception as e:
            logger.exception(e)
            # 只统计转换出错的数量
            if "Convert failed" in str(e):
                total_convert_error += 1
            elif "Convert timeout" in str(e):
                total_time_out += 1
            elif "File is broken" in str(e):
                total_file_broken += 1
            elif "Unsupported file format" in str(e):
                total_unsupported_files += 1
            # abort(f'Error: {traceback.format_exc()}')

    if not input_file_path:
        logger.error(f"Error: Missing argument '--file-path'.")
        abort(f"Error: Missing argument '--file-path'.")
    else:
        '''适配多个文档的list文件输入'''
        if input_file_path.endswith(".list"):
            with open(input_file_path, "r") as f:
                for line in f.readlines():
                    line = line.strip()
                    parse_doc(line, progress_file_path)
        else:
            '''适配单个文档的输入'''
            parse_doc(input_file_path, progress_file_path)

    logger.info(f"total cost time: {int(total_cost_time)} seconds")
    logger.info(f"total convert error: {total_convert_error}")
    logger.info(f"total file broken: {total_file_broken}")
    logger.info(f"total unsupported files: {total_unsupported_files}")
    logger.info(f"total time out: {total_time_out}")
    logger.info(f"total success: {total_success_convert}")


parse_pdf_methods = click.Choice(["ocr", "digital", "auto"])
@cli.command()
@click.option("-m", "--method", "method", type=parse_pdf_methods,
    help="指定解析方法。digital: 文本型 pdf 解析方法， ocr: 光学识别解析 pdf, auto: 程序智能选择解析方法",
    default="auto")
@click.option('-f', '--file-path', 'doc_path', type=click.STRING,
              help='file path, support local or s3')
@click.option("-o", "--output", "output", default="", type=click.STRING)
@click.option("-d", "--debug", is_flag=True, default=False)
@click.option("--models-dir", "models_dir", default="/tmp/models/", type=click.STRING, help="specify the models path")
def pdf_cli(method, doc_path, output, debug, models_dir):

    os.environ["APPLY_FORMULA"] = "TRUE"

    if output_dir == "":
        if os.path.isdir(doc_path):
            output_dir = os.path.join(doc_path, "output")
        else:
            output_dir = os.path.join(os.path.dirname(output), "output")
    
    from magic_pdf.tools.common import do_parse
    from magic_pdf.libs.MakeContentConfig import MakeMode
    model = SingletonModelWrapper() 

    if doc_path.startswith("s3://"):
        bucket, key = parse_s3path(doc_path)
        ak, sk, endpoint = get_s3_config(bucket)
        s3_config = S3Config(ak, sk, endpoint)
    else:
        '''非s3路径不需要初始化s3配置'''
        s3_config = None
    doc_conv = DocConverter(s3_config, models_dir=models_dir)
    bits = doc_conv.get_raw_file_content(doc_path)

    file_name = str(Path(doc_path).stem)
    model_list = model(bits)

    if method == "digital":
        method = "txt"
    do_parse(
        output_dir,
        file_name, 
        bits, 
        model_list, 
        method, 
        f_draw_span_bbox=True,
        f_draw_layout_bbox=True,
        f_dump_md=True,
        f_dump_middle_json=True,
        f_dump_model_json=True,
        f_dump_orig_pdf=True,
        f_dump_content_list=False,
        f_make_md_mode=MakeMode.MM_MD,
        f_draw_model_bbox=False,
    )


if __name__ == '__main__':
    cli()



import logging
import time

import torch
import yaml

from magic_pdf.model.pdf_extract_kit import CustomPEKModel
from magic_doc.utils.yaml_load import patch_yaml_load_with_env

logging.disable(logging.WARNING)


class DocAnalysis(object):
    """
    Description:
      class definition of DocAnalysis module:
    """

    def __init__(
        self, 
        models_dir = '/tmp/models',
        configs="magic_doc/resources/model/model_configs.yaml", 
        **kwargs
    ):
        """
        Description:
          initialize the class instance

        Parameters:
          configs: path to config that contains model weights.
          apply_layout: do layout analysis or not, must be True (defaults to be True).
          apply_formula: do formulat detection and recognition or not, defaults to be False.
          apply_ocr: do ocr(text detection and recognition) or not, defaults to be False.

        """
        self.configs = patch_yaml_load_with_env(configs, "model", yaml.FullLoader)  # load config and patch with env var !
        model_input = {
            "ocr": kwargs.get("apply_ocr", self.configs["models"]["ocr"]),
            "show_log": True,
            "layout": kwargs.get("apply_layout", self.configs["models"]["layout"]), 
            "formula": kwargs.get("apply_formula", self.configs["models"]["formula"]), 
            "table_config": kwargs.get("table_config", self.configs["models"]["table_config"]),
            "models_dir": models_dir,
            "device": "cuda" if torch.cuda.is_available() else "cpu",
        }
        self._pdf_extract_model = CustomPEKModel(**model_input)
        
        logging.info("DocAnalysis init done!")

    def __call__(self, image_list):
        """
        Description:
          do document analysis on input images

        Parameters:
          image_list: list of image array, [np.array, ...]

        Return:
          result: doc analysis result
        """
        model_json = []
        doc_analyze_start = time.time()
        for index, img_dict in enumerate(image_list):
            img = img_dict["img"]
            page_width = img_dict["width"]
            page_height = img_dict["height"]
            result = self._pdf_extract_model(img)
            page_info = {"page_no": index, "height": page_height, "width": page_width}
            page_dict = {"layout_dets": result, "page_info": page_info}
            model_json.append(page_dict)
        doc_analyze_cost = time.time() - doc_analyze_start
        logging.info(f"doc analyze cost: {doc_analyze_cost}")

        return model_json



from abc import ABC, abstractmethod
from typing import List, TypeAlias, TypedDict
from werkzeug.datastructures import FileStorage


class Content(TypedDict):
    # 类型: image/text/md
    type: str

    # 数据
    # image: s3路径 s3://doc/xxx.png
    # text: 文本行
    # md: Markdown格式的文本
    data: str


class Page(TypedDict):
    # 从0开始
    page_no: int

    # 内容列表
    content_list: List[Content]


ExtractResponse: TypeAlias = List[Page]


if __name__ == "__main__":
    pages_data: ExtractResponse = [
        {
            "page_no": 0,
            "content_list": [
                {
                    "type": "text",
                    "data": "This is some text content.",
                },
                {
                    "type": "image",
                    "data": "s3://somebucket/imagepath.jpg",
                },
            ],
        }
    ]


class Extractor(ABC):
    @abstractmethod
    def setup(self):
        pass

    @abstractmethod
    def run(self, file_parse_id: str, r: FileStorage, skip_image: bool = True) -> ExtractResponse:
        pass



import random
import fitz

from magic_doc.contrib.model import (
    ExtractResponse,
    Extractor,
    Page,
    Content,
)
from magic_doc.contrib.wrapper_exceptions import NotSupportOcrPDFException

from werkzeug.datastructures import FileStorage
from loguru import logger


class PDFExtractor(Extractor):
    def __init__(self) -> None:
        super().__init__()

    def setup(self):
        pass

    def is_digital(self, doc, check_page=10, text_len_thrs=100):
        sample_page_num = min(check_page, doc.page_count)
        page_ids = random.sample(range(doc.page_count), sample_page_num)
        page_text_len = [
            len(doc[pno].get_text("text")) > text_len_thrs for pno in page_ids
        ]
        if any(page_text_len):
            return True
        return False

    # Guess kimi implemetation
    def get_text_with_pymupdf(self, doc):
        pages = []
        page_no = 0
        for page in doc:
            content_list = []
            for block in page.get_text("blocks"):
                x0, y0, x1, y1, block_text, block_no, block_type = block
                lf_count = 0
                for ch in block_text:
                    if ch == "\n":
                        lf_count += 1
                block_text = (
                    block_text.replace("-\n", "")
                    .replace("´\n", "´")
                    .replace(" \n", " ")
                )
                if lf_count >= 2:
                    block_text = block_text.replace("\n", " ").strip()
                if len(block_text.strip()) == 0:
                    continue
                content_list.append(
                    Content(
                        type="text",
                        data=block_text,
                    )
                )
            pages.append(Page(page_no=page_no, content_list=content_list))
            page_no += 1
        return pages

    def run(
        self, file_parse_id: str, r: FileStorage, skip_image: bool = True
    ) -> ExtractResponse:
        file_content = r.stream.read()
        with fitz.open(stream=file_content) as doc:
            if self.is_digital(doc):
                logger.info(f"{file_parse_id} is digital pdf")
                return self.get_text_with_pymupdf(doc)
        raise NotSupportOcrPDFException


if __name__ == "__main__":
    pdf_extractor = PDFExtractor()
    with open("magic_doc/contrib/test_data/pdf/test.pdf", "rb") as f:
        logger.info(pdf_extractor.run("test", FileStorage(f, filename="STL.pdf")))



import re

def layout_rm_equation(layout_res):
    rm_idxs = []
    for idx, ele in enumerate(layout_res['layout_dets']):
        if ele['category_id'] == 10:
            rm_idxs.append(idx)
    
    for idx in rm_idxs[::-1]:
        del layout_res['layout_dets'][idx]
    return layout_res


def get_croped_image(image_pil, bbox):
    x_min, y_min, x_max, y_max = bbox
    croped_img = image_pil.crop((x_min, y_min, x_max, y_max))
    return croped_img


def latex_rm_whitespace(s: str):
    """Remove unnecessary whitespace from LaTeX code.
    """
    text_reg = r'(\\(operatorname|mathrm|text|mathbf)\s?\*? {.*?})'
    letter = '[a-zA-Z]'
    noletter = '[\W_^\d]'
    names = [x[0].replace(' ', '') for x in re.findall(text_reg, s)]
    s = re.sub(text_reg, lambda match: str(names.pop(0)), s)
    news = s
    while True:
        s = news
        news = re.sub(r'(?!\\ )(%s)\s+?(%s)' % (noletter, noletter), r'\1\2', s)
        news = re.sub(r'(?!\\ )(%s)\s+?(%s)' % (noletter, letter), r'\1\2', news)
        news = re.sub(r'(%s)\s+?(%s)' % (letter, noletter), r'\1\2', news)
        if news == s:
            break
    return s


# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""PyTorch LayoutLMv3 model. """
import math

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.checkpoint
from torch.nn import BCEWithLogitsLoss, CrossEntropyLoss, MSELoss

from transformers import apply_chunking_to_forward
from transformers.modeling_outputs import (
    BaseModelOutputWithPastAndCrossAttentions,
    BaseModelOutputWithPoolingAndCrossAttentions,
    MaskedLMOutput,
    TokenClassifierOutput,
    QuestionAnsweringModelOutput,
    SequenceClassifierOutput,
)
from transformers.modeling_utils import PreTrainedModel, find_pruneable_heads_and_indices, prune_linear_layer
from transformers.models.roberta.modeling_roberta import (
    RobertaIntermediate,
    RobertaLMHead,
    RobertaOutput,
    RobertaSelfOutput,
)
from transformers.utils import logging

from .configuration_layoutlmv3 import LayoutLMv3Config
from timm.models.layers import to_2tuple


logger = logging.get_logger(__name__)


class PatchEmbed(nn.Module):
    """ Image to Patch Embedding
    """
    def __init__(self, img_size=224, patch_size=16, in_chans=3, embed_dim=768):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)
        self.patch_shape = (img_size[0] // patch_size[0], img_size[1] // patch_size[1])
        self.proj = nn.Conv2d(in_chans, embed_dim, kernel_size=patch_size, stride=patch_size)
        # The following variables are used in detection mycheckpointer.py
        self.num_patches = (img_size[1] // patch_size[1]) * (img_size[0] // patch_size[0])
        self.num_patches_w = self.patch_shape[0]
        self.num_patches_h = self.patch_shape[1]

    def forward(self, x, position_embedding=None):
        x = self.proj(x)

        if position_embedding is not None:
            # interpolate the position embedding to the corresponding size
            position_embedding = position_embedding.view(1, self.patch_shape[0], self.patch_shape[1], -1).permute(0, 3, 1, 2)
            Hp, Wp = x.shape[2], x.shape[3]
            position_embedding = F.interpolate(position_embedding, size=(Hp, Wp), mode='bicubic')
            x = x + position_embedding

        x = x.flatten(2).transpose(1, 2)
        return x

class LayoutLMv3Embeddings(nn.Module):
    """
    Same as BertEmbeddings with a tiny tweak for positional embeddings indexing.
    """

    # Copied from transformers.models.bert.modeling_bert.BertEmbeddings.__init__
    def __init__(self, config):
        super().__init__()
        self.word_embeddings = nn.Embedding(config.vocab_size, config.hidden_size, padding_idx=config.pad_token_id)
        self.token_type_embeddings = nn.Embedding(config.type_vocab_size, config.hidden_size)

        self.LayerNorm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

        # position_ids (1, len position emb) is contiguous in memory and exported when serialized
        self.register_buffer("position_ids", torch.arange(config.max_position_embeddings).expand((1, -1)))

        # End copy
        self.padding_idx = config.pad_token_id
        self.position_embeddings = nn.Embedding(
            config.max_position_embeddings, config.hidden_size, padding_idx=self.padding_idx
        )

        self.x_position_embeddings = nn.Embedding(config.max_2d_position_embeddings, config.coordinate_size)
        self.y_position_embeddings = nn.Embedding(config.max_2d_position_embeddings, config.coordinate_size)
        self.h_position_embeddings = nn.Embedding(config.max_2d_position_embeddings, config.shape_size)
        self.w_position_embeddings = nn.Embedding(config.max_2d_position_embeddings, config.shape_size)

    def _calc_spatial_position_embeddings(self, bbox):
        try:
            assert torch.all(0 <= bbox) and torch.all(bbox <= 1023)
            left_position_embeddings = self.x_position_embeddings(bbox[:, :, 0])
            upper_position_embeddings = self.y_position_embeddings(bbox[:, :, 1])
            right_position_embeddings = self.x_position_embeddings(bbox[:, :, 2])
            lower_position_embeddings = self.y_position_embeddings(bbox[:, :, 3])
        except IndexError as e:
            raise IndexError("The :obj:`bbox` coordinate values should be within 0-1000 range.") from e

        h_position_embeddings = self.h_position_embeddings(torch.clip(bbox[:, :, 3] - bbox[:, :, 1], 0, 1023))
        w_position_embeddings = self.w_position_embeddings(torch.clip(bbox[:, :, 2] - bbox[:, :, 0], 0, 1023))

        # below is the difference between LayoutLMEmbeddingsV2 (torch.cat) and LayoutLMEmbeddingsV1 (add)
        spatial_position_embeddings = torch.cat(
            [
                left_position_embeddings,
                upper_position_embeddings,
                right_position_embeddings,
                lower_position_embeddings,
                h_position_embeddings,
                w_position_embeddings,
            ],
            dim=-1,
        )
        return spatial_position_embeddings

    def create_position_ids_from_input_ids(self, input_ids, padding_idx, past_key_values_length=0):
        """
        Replace non-padding symbols with their position numbers. Position numbers begin at padding_idx+1. Padding symbols
        are ignored. This is modified from fairseq's `utils.make_positions`.

        Args:
            x: torch.Tensor x:

        Returns: torch.Tensor
        """
        # The series of casts and type-conversions here are carefully balanced to both work with ONNX export and XLA.
        mask = input_ids.ne(padding_idx).int()
        incremental_indices = (torch.cumsum(mask, dim=1).type_as(mask) + past_key_values_length) * mask
        return incremental_indices.long() + padding_idx

    def forward(
        self,
        input_ids=None,
        bbox=None,
        token_type_ids=None,
        position_ids=None,
        inputs_embeds=None,
        past_key_values_length=0,
    ):
        if position_ids is None:
            if input_ids is not None:
                # Create the position ids from the input token ids. Any padded tokens remain padded.
                position_ids = self.create_position_ids_from_input_ids(
                    input_ids, self.padding_idx, past_key_values_length).to(input_ids.device)
            else:
                position_ids = self.create_position_ids_from_inputs_embeds(inputs_embeds)

        if input_ids is not None:
            input_shape = input_ids.size()
        else:
            input_shape = inputs_embeds.size()[:-1]

        if token_type_ids is None:
            token_type_ids = torch.zeros(input_shape, dtype=torch.long, device=self.position_ids.device)

        if inputs_embeds is None:
            inputs_embeds = self.word_embeddings(input_ids)
        token_type_embeddings = self.token_type_embeddings(token_type_ids)

        embeddings = inputs_embeds + token_type_embeddings
        position_embeddings = self.position_embeddings(position_ids)
        embeddings += position_embeddings

        spatial_position_embeddings = self._calc_spatial_position_embeddings(bbox)

        embeddings = embeddings + spatial_position_embeddings

        embeddings = self.LayerNorm(embeddings)
        embeddings = self.dropout(embeddings)
        return embeddings

    def create_position_ids_from_inputs_embeds(self, inputs_embeds):
        """
        We are provided embeddings directly. We cannot infer which are padded so just generate sequential position ids.

        Args:
            inputs_embeds: torch.Tensor≈

        Returns: torch.Tensor
        """
        input_shape = inputs_embeds.size()[:-1]
        sequence_length = input_shape[1]

        position_ids = torch.arange(
            self.padding_idx + 1, sequence_length + self.padding_idx + 1, dtype=torch.long, device=inputs_embeds.device
        )
        return position_ids.unsqueeze(0).expand(input_shape)


class LayoutLMv3PreTrainedModel(PreTrainedModel):
    """
    An abstract class to handle weights initialization and a simple interface for downloading and loading pretrained
    models.
    """

    config_class = LayoutLMv3Config
    base_model_prefix = "layoutlmv3"

    # Copied from transformers.models.bert.modeling_bert.BertPreTrainedModel._init_weights
    def _init_weights(self, module):
        """Initialize the weights"""
        if isinstance(module, nn.Linear):
            # Slightly different from the TF version which uses truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)


class LayoutLMv3SelfAttention(nn.Module):
    def __init__(self, config):
        super().__init__()
        if config.hidden_size % config.num_attention_heads != 0 and not hasattr(config, "embedding_size"):
            raise ValueError(
                f"The hidden size ({config.hidden_size}) is not a multiple of the number of attention "
                f"heads ({config.num_attention_heads})"
            )

        self.num_attention_heads = config.num_attention_heads
        self.attention_head_size = int(config.hidden_size / config.num_attention_heads)
        self.all_head_size = self.num_attention_heads * self.attention_head_size

        self.query = nn.Linear(config.hidden_size, self.all_head_size)
        self.key = nn.Linear(config.hidden_size, self.all_head_size)
        self.value = nn.Linear(config.hidden_size, self.all_head_size)

        self.dropout = nn.Dropout(config.attention_probs_dropout_prob)
        self.has_relative_attention_bias = config.has_relative_attention_bias
        self.has_spatial_attention_bias = config.has_spatial_attention_bias

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (self.num_attention_heads, self.attention_head_size)
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)

    def cogview_attn(self, attention_scores, alpha=32):
        '''
        https://arxiv.org/pdf/2105.13290.pdf
        Section 2.4 Stabilization of training: Precision Bottleneck Relaxation (PB-Relax).
        A replacement of the original nn.Softmax(dim=-1)(attention_scores)
        Seems the new attention_probs will result in a slower speed and a little bias
        Can use torch.allclose(standard_attention_probs, cogview_attention_probs, atol=1e-08) for comparison
        The smaller atol (e.g., 1e-08), the better.
        '''
        scaled_attention_scores = attention_scores / alpha
        max_value = scaled_attention_scores.amax(dim=(-1)).unsqueeze(-1)
        # max_value = scaled_attention_scores.amax(dim=(-2, -1)).unsqueeze(-1).unsqueeze(-1)
        new_attention_scores = (scaled_attention_scores - max_value) * alpha
        return nn.Softmax(dim=-1)(new_attention_scores)

    def forward(
        self,
        hidden_states,
        attention_mask=None,
        head_mask=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        past_key_value=None,
        output_attentions=False,
        rel_pos=None,
        rel_2d_pos=None,
    ):
        mixed_query_layer = self.query(hidden_states)

        # If this is instantiated as a cross-attention module, the keys
        # and values come from an encoder; the attention mask needs to be
        # such that the encoder's padding tokens are not attended to.
        is_cross_attention = encoder_hidden_states is not None

        if is_cross_attention and past_key_value is not None:
            # reuse k,v, cross_attentions
            key_layer = past_key_value[0]
            value_layer = past_key_value[1]
            attention_mask = encoder_attention_mask
        elif is_cross_attention:
            key_layer = self.transpose_for_scores(self.key(encoder_hidden_states))
            value_layer = self.transpose_for_scores(self.value(encoder_hidden_states))
            attention_mask = encoder_attention_mask
        elif past_key_value is not None:
            key_layer = self.transpose_for_scores(self.key(hidden_states))
            value_layer = self.transpose_for_scores(self.value(hidden_states))
            key_layer = torch.cat([past_key_value[0], key_layer], dim=2)
            value_layer = torch.cat([past_key_value[1], value_layer], dim=2)
        else:
            key_layer = self.transpose_for_scores(self.key(hidden_states))
            value_layer = self.transpose_for_scores(self.value(hidden_states))

        query_layer = self.transpose_for_scores(mixed_query_layer)

        # Take the dot product between "query" and "key" to get the raw attention scores.
        # The attention scores QT K/√d could be significantly larger than input elements, and result in overflow.
        # Changing the computational order into QT(K/√d) alleviates the problem. (https://arxiv.org/pdf/2105.13290.pdf)
        attention_scores = torch.matmul(query_layer / math.sqrt(self.attention_head_size), key_layer.transpose(-1, -2))

        if self.has_relative_attention_bias and self.has_spatial_attention_bias:
            attention_scores += (rel_pos + rel_2d_pos) / math.sqrt(self.attention_head_size)
        elif self.has_relative_attention_bias:
            attention_scores += rel_pos / math.sqrt(self.attention_head_size)

        # if self.has_relative_attention_bias:
        #     attention_scores += rel_pos
        # if self.has_spatial_attention_bias:
        #     attention_scores += rel_2d_pos

        # attention_scores = attention_scores / math.sqrt(self.attention_head_size)
        if attention_mask is not None:
            # Apply the attention mask is (precomputed for all layers in RobertaModel forward() function)
            attention_scores = attention_scores + attention_mask

        # Normalize the attention scores to probabilities.
        # attention_probs = nn.Softmax(dim=-1)(attention_scores)  # comment the line below and use this line for speedup
        attention_probs = self.cogview_attn(attention_scores)  # to stablize training
        # assert torch.allclose(attention_probs, nn.Softmax(dim=-1)(attention_scores), atol=1e-8)

        # This is actually dropping out entire tokens to attend to, which might
        # seem a bit unusual, but is taken from the original Transformer paper.
        attention_probs = self.dropout(attention_probs)

        # Mask heads if we want to
        if head_mask is not None:
            attention_probs = attention_probs * head_mask

        context_layer = torch.matmul(attention_probs, value_layer)

        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
        new_context_layer_shape = context_layer.size()[:-2] + (self.all_head_size,)
        context_layer = context_layer.view(*new_context_layer_shape)

        outputs = (context_layer, attention_probs) if output_attentions else (context_layer,)

        return outputs


class LayoutLMv3Attention(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.self = LayoutLMv3SelfAttention(config)
        self.output = RobertaSelfOutput(config)
        self.pruned_heads = set()

    def prune_heads(self, heads):
        if len(heads) == 0:
            return
        heads, index = find_pruneable_heads_and_indices(
            heads, self.self.num_attention_heads, self.self.attention_head_size, self.pruned_heads
        )

        # Prune linear layers
        self.self.query = prune_linear_layer(self.self.query, index)
        self.self.key = prune_linear_layer(self.self.key, index)
        self.self.value = prune_linear_layer(self.self.value, index)
        self.output.dense = prune_linear_layer(self.output.dense, index, dim=1)

        # Update hyper params and store pruned heads
        self.self.num_attention_heads = self.self.num_attention_heads - len(heads)
        self.self.all_head_size = self.self.attention_head_size * self.self.num_attention_heads
        self.pruned_heads = self.pruned_heads.union(heads)

    def forward(
        self,
        hidden_states,
        attention_mask=None,
        head_mask=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        past_key_value=None,
        output_attentions=False,
        rel_pos=None,
        rel_2d_pos=None,
    ):
        self_outputs = self.self(
            hidden_states,
            attention_mask,
            head_mask,
            encoder_hidden_states,
            encoder_attention_mask,
            past_key_value,
            output_attentions,
            rel_pos=rel_pos,
            rel_2d_pos=rel_2d_pos,
        )
        attention_output = self.output(self_outputs[0], hidden_states)
        outputs = (attention_output,) + self_outputs[1:]  # add attentions if we output them
        return outputs


class LayoutLMv3Layer(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.chunk_size_feed_forward = config.chunk_size_feed_forward
        self.seq_len_dim = 1
        self.attention = LayoutLMv3Attention(config)
        assert not config.is_decoder and not config.add_cross_attention, \
            "This version do not support decoder. Please refer to RoBERTa for implementation of is_decoder."
        self.intermediate = RobertaIntermediate(config)
        self.output = RobertaOutput(config)

    def forward(
        self,
        hidden_states,
        attention_mask=None,
        head_mask=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        past_key_value=None,
        output_attentions=False,
        rel_pos=None,
        rel_2d_pos=None,
    ):
        # decoder uni-directional self-attention cached key/values tuple is at positions 1,2
        self_attn_past_key_value = past_key_value[:2] if past_key_value is not None else None
        self_attention_outputs = self.attention(
            hidden_states,
            attention_mask,
            head_mask,
            output_attentions=output_attentions,
            past_key_value=self_attn_past_key_value,
            rel_pos=rel_pos,
            rel_2d_pos=rel_2d_pos,
        )
        attention_output = self_attention_outputs[0]

        outputs = self_attention_outputs[1:]  # add self attentions if we output attention weights

        layer_output = apply_chunking_to_forward(
            self.feed_forward_chunk, self.chunk_size_feed_forward, self.seq_len_dim, attention_output
        )
        outputs = (layer_output,) + outputs

        return outputs

    def feed_forward_chunk(self, attention_output):
        intermediate_output = self.intermediate(attention_output)
        layer_output = self.output(intermediate_output, attention_output)
        return layer_output


class LayoutLMv3Encoder(nn.Module):
    def __init__(self, config, detection=False, out_features=None):
        super().__init__()
        self.config = config
        self.detection = detection
        self.layer = nn.ModuleList([LayoutLMv3Layer(config) for _ in range(config.num_hidden_layers)])
        self.gradient_checkpointing = False

        self.has_relative_attention_bias = config.has_relative_attention_bias
        self.has_spatial_attention_bias = config.has_spatial_attention_bias

        if self.has_relative_attention_bias:
            self.rel_pos_bins = config.rel_pos_bins
            self.max_rel_pos = config.max_rel_pos
            self.rel_pos_onehot_size = config.rel_pos_bins
            self.rel_pos_bias = nn.Linear(self.rel_pos_onehot_size, config.num_attention_heads, bias=False)

        if self.has_spatial_attention_bias:
            self.max_rel_2d_pos = config.max_rel_2d_pos
            self.rel_2d_pos_bins = config.rel_2d_pos_bins
            self.rel_2d_pos_onehot_size = config.rel_2d_pos_bins
            self.rel_pos_x_bias = nn.Linear(self.rel_2d_pos_onehot_size, config.num_attention_heads, bias=False)
            self.rel_pos_y_bias = nn.Linear(self.rel_2d_pos_onehot_size, config.num_attention_heads, bias=False)

        if self.detection:
            self.gradient_checkpointing = True
            embed_dim = self.config.hidden_size
            self.out_features = out_features
            self.out_indices = [int(name[5:]) for name in out_features]
            self.fpn1 = nn.Sequential(
                nn.ConvTranspose2d(embed_dim, embed_dim, kernel_size=2, stride=2),
                # nn.SyncBatchNorm(embed_dim),
                nn.BatchNorm2d(embed_dim),
                nn.GELU(),
                nn.ConvTranspose2d(embed_dim, embed_dim, kernel_size=2, stride=2),
            )

            self.fpn2 = nn.Sequential(
                nn.ConvTranspose2d(embed_dim, embed_dim, kernel_size=2, stride=2),
            )

            self.fpn3 = nn.Identity()

            self.fpn4 = nn.MaxPool2d(kernel_size=2, stride=2)
            self.ops = [self.fpn1, self.fpn2, self.fpn3, self.fpn4]

    def relative_position_bucket(self, relative_position, bidirectional=True, num_buckets=32, max_distance=128):
        ret = 0
        if bidirectional:
            num_buckets //= 2
            ret += (relative_position > 0).long() * num_buckets
            n = torch.abs(relative_position)
        else:
            n = torch.max(-relative_position, torch.zeros_like(relative_position))
        # now n is in the range [0, inf)

        # half of the buckets are for exact increments in positions
        max_exact = num_buckets // 2
        is_small = n < max_exact

        # The other half of the buckets are for logarithmically bigger bins in positions up to max_distance
        val_if_large = max_exact + (
                torch.log(n.float() / max_exact) / math.log(max_distance / max_exact) * (num_buckets - max_exact)
        ).to(torch.long)
        val_if_large = torch.min(val_if_large, torch.full_like(val_if_large, num_buckets - 1))

        ret += torch.where(is_small, n, val_if_large)
        return ret

    def _cal_1d_pos_emb(self, hidden_states, position_ids, valid_span):
        VISUAL_NUM = 196 + 1

        rel_pos_mat = position_ids.unsqueeze(-2) - position_ids.unsqueeze(-1)

        if valid_span is not None:
            # for the text part, if two words are not in the same line,
            # set their distance to the max value (position_ids.shape[-1])
            rel_pos_mat[(rel_pos_mat > 0) & (valid_span == False)] = position_ids.shape[1]
            rel_pos_mat[(rel_pos_mat < 0) & (valid_span == False)] = -position_ids.shape[1]

            # image-text, minimum distance
            rel_pos_mat[:, -VISUAL_NUM:, :-VISUAL_NUM] = 0
            rel_pos_mat[:, :-VISUAL_NUM, -VISUAL_NUM:] = 0

        rel_pos = self.relative_position_bucket(
            rel_pos_mat,
            num_buckets=self.rel_pos_bins,
            max_distance=self.max_rel_pos,
        )
        rel_pos = F.one_hot(rel_pos, num_classes=self.rel_pos_onehot_size).type_as(hidden_states)
        rel_pos = self.rel_pos_bias(rel_pos).permute(0, 3, 1, 2)
        rel_pos = rel_pos.contiguous()
        return rel_pos

    def _cal_2d_pos_emb(self, hidden_states, bbox):
        position_coord_x = bbox[:, :, 0]
        position_coord_y = bbox[:, :, 3]
        rel_pos_x_2d_mat = position_coord_x.unsqueeze(-2) - position_coord_x.unsqueeze(-1)
        rel_pos_y_2d_mat = position_coord_y.unsqueeze(-2) - position_coord_y.unsqueeze(-1)
        rel_pos_x = self.relative_position_bucket(
            rel_pos_x_2d_mat,
            num_buckets=self.rel_2d_pos_bins,
            max_distance=self.max_rel_2d_pos,
        )
        rel_pos_y = self.relative_position_bucket(
            rel_pos_y_2d_mat,
            num_buckets=self.rel_2d_pos_bins,
            max_distance=self.max_rel_2d_pos,
        )
        rel_pos_x = F.one_hot(rel_pos_x, num_classes=self.rel_2d_pos_onehot_size).type_as(hidden_states)
        rel_pos_y = F.one_hot(rel_pos_y, num_classes=self.rel_2d_pos_onehot_size).type_as(hidden_states)
        rel_pos_x = self.rel_pos_x_bias(rel_pos_x).permute(0, 3, 1, 2)
        rel_pos_y = self.rel_pos_y_bias(rel_pos_y).permute(0, 3, 1, 2)
        rel_pos_x = rel_pos_x.contiguous()
        rel_pos_y = rel_pos_y.contiguous()
        rel_2d_pos = rel_pos_x + rel_pos_y
        return rel_2d_pos

    def forward(
        self,
        hidden_states,
        bbox=None,
        attention_mask=None,
        head_mask=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        past_key_values=None,
        use_cache=None,
        output_attentions=False,
        output_hidden_states=False,
        return_dict=True,
        position_ids=None,
        Hp=None,
        Wp=None,
        valid_span=None,
    ):
        all_hidden_states = () if output_hidden_states else None
        all_self_attentions = () if output_attentions else None
        all_cross_attentions = () if output_attentions and self.config.add_cross_attention else None

        next_decoder_cache = () if use_cache else None

        rel_pos = self._cal_1d_pos_emb(hidden_states, position_ids, valid_span) if self.has_relative_attention_bias else None
        rel_2d_pos = self._cal_2d_pos_emb(hidden_states, bbox) if self.has_spatial_attention_bias else None

        if self.detection:
            feat_out = {}
            j = 0

        for i, layer_module in enumerate(self.layer):
            if output_hidden_states:
                all_hidden_states = all_hidden_states + (hidden_states,)

            layer_head_mask = head_mask[i] if head_mask is not None else None
            past_key_value = past_key_values[i] if past_key_values is not None else None

            if self.gradient_checkpointing and self.training:

                if use_cache:
                    logger.warning(
                        "`use_cache=True` is incompatible with gradient checkpointing. Setting `use_cache=False`..."
                    )
                    use_cache = False

                def create_custom_forward(module):
                    def custom_forward(*inputs):
                        return module(*inputs)
                        # return module(*inputs, past_key_value, output_attentions, rel_pos, rel_2d_pos)
                        # The above line will cause error:
                        # RuntimeError: Trying to backward through the graph a second time
                        # (or directly access saved tensors after they have already been freed).
                    return custom_forward

                layer_outputs = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(layer_module),
                    hidden_states,
                    attention_mask,
                    layer_head_mask,
                    encoder_hidden_states,
                    encoder_attention_mask,
                    past_key_value,
                    output_attentions,
                    rel_pos,
                    rel_2d_pos
                )
            else:
                layer_outputs = layer_module(
                    hidden_states,
                    attention_mask,
                    layer_head_mask,
                    encoder_hidden_states,
                    encoder_attention_mask,
                    past_key_value,
                    output_attentions,
                    rel_pos=rel_pos,
                    rel_2d_pos=rel_2d_pos,
                )

            hidden_states = layer_outputs[0]
            if use_cache:
                next_decoder_cache += (layer_outputs[-1],)
            if output_attentions:
                all_self_attentions = all_self_attentions + (layer_outputs[1],)
                if self.config.add_cross_attention:
                    all_cross_attentions = all_cross_attentions + (layer_outputs[2],)

            if self.detection and i in self.out_indices:
                xp = hidden_states[:, -Hp*Wp:, :].permute(0, 2, 1).reshape(len(hidden_states), -1, Hp, Wp)
                feat_out[self.out_features[j]] = self.ops[j](xp.contiguous())
                j += 1

        if self.detection:
            return feat_out

        if output_hidden_states:
            all_hidden_states = all_hidden_states + (hidden_states,)

        if not return_dict:
            return tuple(
                v
                for v in [
                    hidden_states,
                    next_decoder_cache,
                    all_hidden_states,
                    all_self_attentions,
                    all_cross_attentions,
                ]
                if v is not None
            )
        return BaseModelOutputWithPastAndCrossAttentions(
            last_hidden_state=hidden_states,
            past_key_values=next_decoder_cache,
            hidden_states=all_hidden_states,
            attentions=all_self_attentions,
            cross_attentions=all_cross_attentions,
        )


class LayoutLMv3Model(LayoutLMv3PreTrainedModel):
    """
    """

    _keys_to_ignore_on_load_missing = [r"position_ids"]

    # Copied from transformers.models.bert.modeling_bert.BertModel.__init__ with Bert->Roberta
    def __init__(self, config, detection=False, out_features=None, image_only=False):
        super().__init__(config)
        self.config = config
        assert not config.is_decoder and not config.add_cross_attention, \
            "This version do not support decoder. Please refer to RoBERTa for implementation of is_decoder."
        self.detection = detection
        if not self.detection:
            self.image_only = False
        else:
            assert config.visual_embed
            self.image_only = image_only

        if not self.image_only:
            self.embeddings = LayoutLMv3Embeddings(config)
        self.encoder = LayoutLMv3Encoder(config, detection=detection, out_features=out_features)

        if config.visual_embed:
            embed_dim = self.config.hidden_size
            # use the default pre-training parameters for fine-tuning (e.g., input_size)
            # when the input_size is larger in fine-tuning, we will interpolate the position embedding in forward
            self.patch_embed = PatchEmbed(embed_dim=embed_dim)

            patch_size = 16
            size = int(self.config.input_size / patch_size)
            self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
            self.pos_embed = nn.Parameter(torch.zeros(1, size * size + 1, embed_dim))
            self.pos_drop = nn.Dropout(p=0.)

            self.LayerNorm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
            self.dropout = nn.Dropout(config.hidden_dropout_prob)

            if self.config.has_relative_attention_bias or self.config.has_spatial_attention_bias:
                self._init_visual_bbox(img_size=(size, size))

            from functools import partial
            norm_layer = partial(nn.LayerNorm, eps=1e-6)
            self.norm = norm_layer(embed_dim)

        self.init_weights()

    def get_input_embeddings(self):
        return self.embeddings.word_embeddings

    def set_input_embeddings(self, value):
        self.embeddings.word_embeddings = value

    def _prune_heads(self, heads_to_prune):
        """
        Prunes heads of the model. heads_to_prune: dict of {layer_num: list of heads to prune in this layer} See base
        class PreTrainedModel
        """
        for layer, heads in heads_to_prune.items():
            self.encoder.layer[layer].attention.prune_heads(heads)

    def _init_visual_bbox(self, img_size=(14, 14), max_len=1000):
        visual_bbox_x = torch.div(torch.arange(0, max_len * (img_size[1] + 1), max_len),
                                  img_size[1], rounding_mode='trunc')
        visual_bbox_y = torch.div(torch.arange(0, max_len * (img_size[0] + 1), max_len),
                                  img_size[0], rounding_mode='trunc')
        visual_bbox = torch.stack(
            [
                visual_bbox_x[:-1].repeat(img_size[0], 1),
                visual_bbox_y[:-1].repeat(img_size[1], 1).transpose(0, 1),
                visual_bbox_x[1:].repeat(img_size[0], 1),
                visual_bbox_y[1:].repeat(img_size[1], 1).transpose(0, 1),
            ],
            dim=-1,
        ).view(-1, 4)

        cls_token_box = torch.tensor([[0 + 1, 0 + 1, max_len - 1, max_len - 1]])
        self.visual_bbox = torch.cat([cls_token_box, visual_bbox], dim=0)

    def _calc_visual_bbox(self, device, dtype, bsz):  # , img_size=(14, 14), max_len=1000):
        visual_bbox = self.visual_bbox.repeat(bsz, 1, 1)
        visual_bbox = visual_bbox.to(device).type(dtype)
        return visual_bbox

    def forward_image(self, x):
        if self.detection:
            x = self.patch_embed(x, self.pos_embed[:, 1:, :] if self.pos_embed is not None else None)
        else:
            x = self.patch_embed(x)
        batch_size, seq_len, _ = x.size()

        cls_tokens = self.cls_token.expand(batch_size, -1, -1)  # stole cls_tokens impl from Phil Wang, thanks
        if self.pos_embed is not None and self.detection:
            cls_tokens = cls_tokens + self.pos_embed[:, :1, :]

        x = torch.cat((cls_tokens, x), dim=1)
        if self.pos_embed is not None and not self.detection:
            x = x + self.pos_embed
        x = self.pos_drop(x)

        x = self.norm(x)
        return x

    # Copied from transformers.models.bert.modeling_bert.BertModel.forward
    def forward(
        self,
        input_ids=None,
        bbox=None,
        attention_mask=None,
        token_type_ids=None,
        valid_span=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        past_key_values=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
        images=None,
    ):
        r"""
        encoder_hidden_states  (:obj:`torch.FloatTensor` of shape :obj:`(batch_size, sequence_length, hidden_size)`, `optional`):
            Sequence of hidden-states at the output of the last layer of the encoder. Used in the cross-attention if
            the model is configured as a decoder.
        encoder_attention_mask (:obj:`torch.FloatTensor` of shape :obj:`(batch_size, sequence_length)`, `optional`):
            Mask to avoid performing attention on the padding token indices of the encoder input. This mask is used in
            the cross-attention if the model is configured as a decoder. Mask values selected in ``[0, 1]``:

            - 1 for tokens that are **not masked**,
            - 0 for tokens that are **masked**.
        past_key_values (:obj:`tuple(tuple(torch.FloatTensor))` of length :obj:`config.n_layers` with each tuple having 4 tensors of shape :obj:`(batch_size, num_heads, sequence_length - 1, embed_size_per_head)`):
            Contains precomputed key and value hidden states of the attention blocks. Can be used to speed up decoding.

            If :obj:`past_key_values` are used, the user can optionally input only the last :obj:`decoder_input_ids`
            (those that don't have their past key value states given to this model) of shape :obj:`(batch_size, 1)`
            instead of all :obj:`decoder_input_ids` of shape :obj:`(batch_size, sequence_length)`.
        use_cache (:obj:`bool`, `optional`):
            If set to :obj:`True`, :obj:`past_key_values` key value states are returned and can be used to speed up
            decoding (see :obj:`past_key_values`).
        """
        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        use_cache = False

        # if input_ids is not None and inputs_embeds is not None:
        #     raise ValueError("You cannot specify both input_ids and inputs_embeds at the same time")
        if input_ids is not None:
            input_shape = input_ids.size()
            batch_size, seq_length = input_shape
            device = input_ids.device
        elif inputs_embeds is not None:
            input_shape = inputs_embeds.size()[:-1]
            batch_size, seq_length = input_shape
            device = inputs_embeds.device
        elif images is not None:
            batch_size = len(images)
            device = images.device
        else:
            raise ValueError("You have to specify either input_ids or inputs_embeds or images")

        if not self.image_only:
            # past_key_values_length
            past_key_values_length = past_key_values[0][0].shape[2] if past_key_values is not None else 0

            if attention_mask is None:
                attention_mask = torch.ones(((batch_size, seq_length + past_key_values_length)), device=device)
            if token_type_ids is None:
                token_type_ids = torch.zeros(input_shape, dtype=torch.long, device=device)

        # We can provide a self-attention mask of dimensions [batch_size, from_seq_length, to_seq_length]
        # ourselves in which case we just need to make it broadcastable to all heads.
        # extended_attention_mask: torch.Tensor = self.get_extended_attention_mask(attention_mask, input_shape, device)

        encoder_extended_attention_mask = None

        # Prepare head mask if needed
        # 1.0 in head_mask indicate we keep the head
        # attention_probs has shape bsz x n_heads x N x N
        # input head_mask has shape [num_heads] or [num_hidden_layers x num_heads]
        # and head_mask is converted to shape [num_hidden_layers x batch x num_heads x seq_length x seq_length]
        head_mask = self.get_head_mask(head_mask, self.config.num_hidden_layers)

        if not self.image_only:
            if bbox is None:
                bbox = torch.zeros(tuple(list(input_shape) + [4]), dtype=torch.long, device=device)

            embedding_output = self.embeddings(
                input_ids=input_ids,
                bbox=bbox,
                position_ids=position_ids,
                token_type_ids=token_type_ids,
                inputs_embeds=inputs_embeds,
                past_key_values_length=past_key_values_length,
            )

        final_bbox = final_position_ids = None
        Hp = Wp = None
        if images is not None:
            patch_size = 16
            Hp, Wp = int(images.shape[2] / patch_size), int(images.shape[3] / patch_size)
            visual_emb = self.forward_image(images)
            if self.detection:
                visual_attention_mask = torch.ones((batch_size, visual_emb.shape[1]), dtype=torch.long, device=device)
                if self.image_only:
                    attention_mask = visual_attention_mask
                else:
                    attention_mask = torch.cat([attention_mask, visual_attention_mask], dim=1)
            elif self.image_only:
                attention_mask = torch.ones((batch_size, visual_emb.shape[1]), dtype=torch.long, device=device)

            if self.config.has_relative_attention_bias or self.config.has_spatial_attention_bias:
                if self.config.has_spatial_attention_bias:
                    visual_bbox = self._calc_visual_bbox(device, dtype=torch.long, bsz=batch_size)
                    if self.image_only:
                        final_bbox = visual_bbox
                    else:
                        final_bbox = torch.cat([bbox, visual_bbox], dim=1)

                visual_position_ids = torch.arange(0, visual_emb.shape[1], dtype=torch.long, device=device).repeat(
                    batch_size, 1)
                if self.image_only:
                    final_position_ids = visual_position_ids
                else:
                    position_ids = torch.arange(0, input_shape[1], device=device).unsqueeze(0)
                    position_ids = position_ids.expand_as(input_ids)
                    final_position_ids = torch.cat([position_ids, visual_position_ids], dim=1)

            if self.image_only:
                embedding_output = visual_emb
            else:
                embedding_output = torch.cat([embedding_output, visual_emb], dim=1)
            embedding_output = self.LayerNorm(embedding_output)
            embedding_output = self.dropout(embedding_output)
        elif self.config.has_relative_attention_bias or self.config.has_spatial_attention_bias:
            if self.config.has_spatial_attention_bias:
                final_bbox = bbox
            if self.config.has_relative_attention_bias:
                position_ids = self.embeddings.position_ids[:, :input_shape[1]]
                position_ids = position_ids.expand_as(input_ids)
                final_position_ids = position_ids

        extended_attention_mask: torch.Tensor = self.get_extended_attention_mask(attention_mask, None, device)

        encoder_outputs = self.encoder(
            embedding_output,
            bbox=final_bbox,
            position_ids=final_position_ids,
            attention_mask=extended_attention_mask,
            head_mask=head_mask,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_extended_attention_mask,
            past_key_values=past_key_values,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            Hp=Hp,
            Wp=Wp,
            valid_span=valid_span,
        )

        if self.detection:
            return encoder_outputs

        sequence_output = encoder_outputs[0]
        pooled_output = None

        if not return_dict:
            return (sequence_output, pooled_output) + encoder_outputs[1:]

        return BaseModelOutputWithPoolingAndCrossAttentions(
            last_hidden_state=sequence_output,
            pooler_output=pooled_output,
            past_key_values=encoder_outputs.past_key_values,
            hidden_states=encoder_outputs.hidden_states,
            attentions=encoder_outputs.attentions,
            cross_attentions=encoder_outputs.cross_attentions,
        )


class LayoutLMv3ClassificationHead(nn.Module):
    """
    Head for sentence-level classification tasks.
    Reference: RobertaClassificationHead
    """

    def __init__(self, config, pool_feature=False):
        super().__init__()
        self.pool_feature = pool_feature
        if pool_feature:
            self.dense = nn.Linear(config.hidden_size*3, config.hidden_size)
        else:
            self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        classifier_dropout = (
            config.classifier_dropout if config.classifier_dropout is not None else config.hidden_dropout_prob
        )
        self.dropout = nn.Dropout(classifier_dropout)
        self.out_proj = nn.Linear(config.hidden_size, config.num_labels)

    def forward(self, x):
        # x = features[:, 0, :]  # take <s> token (equiv. to [CLS])
        x = self.dropout(x)
        x = self.dense(x)
        x = torch.tanh(x)
        x = self.dropout(x)
        x = self.out_proj(x)
        return x


class LayoutLMv3ForTokenClassification(LayoutLMv3PreTrainedModel):
    _keys_to_ignore_on_load_unexpected = [r"pooler"]
    _keys_to_ignore_on_load_missing = [r"position_ids"]

    def __init__(self, config):
        super().__init__(config)
        self.num_labels = config.num_labels

        self.layoutlmv3 = LayoutLMv3Model(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        if config.num_labels < 10:
            self.classifier = nn.Linear(config.hidden_size, config.num_labels)
        else:
            self.classifier = LayoutLMv3ClassificationHead(config, pool_feature=False)

        self.init_weights()

    def forward(
        self,
        input_ids=None,
        bbox=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        valid_span=None,
        head_mask=None,
        inputs_embeds=None,
        labels=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
        images=None,
    ):
        r"""
        labels (:obj:`torch.LongTensor` of shape :obj:`(batch_size, sequence_length)`, `optional`):
            Labels for computing the token classification loss. Indices should be in ``[0, ..., config.num_labels -
            1]``.
        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        outputs = self.layoutlmv3(
            input_ids,
            bbox=bbox,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            images=images,
            valid_span=valid_span,
        )

        sequence_output = outputs[0]

        sequence_output = self.dropout(sequence_output)
        logits = self.classifier(sequence_output)

        loss = None
        if labels is not None:
            loss_fct = CrossEntropyLoss()
            # Only keep active parts of the loss
            if attention_mask is not None:
                active_loss = attention_mask.view(-1) == 1
                active_logits = logits.view(-1, self.num_labels)
                active_labels = torch.where(
                    active_loss, labels.view(-1), torch.tensor(loss_fct.ignore_index).type_as(labels)
                )
                loss = loss_fct(active_logits, active_labels)
            else:
                loss = loss_fct(logits.view(-1, self.num_labels), labels.view(-1))

        if not return_dict:
            output = (logits,) + outputs[2:]
            return ((loss,) + output) if loss is not None else output

        return TokenClassifierOutput(
            loss=loss,
            logits=logits,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )


class LayoutLMv3ForQuestionAnswering(LayoutLMv3PreTrainedModel):
    _keys_to_ignore_on_load_unexpected = [r"pooler"]
    _keys_to_ignore_on_load_missing = [r"position_ids"]

    def __init__(self, config):
        super().__init__(config)
        self.num_labels = config.num_labels

        self.layoutlmv3 = LayoutLMv3Model(config)
        # self.qa_outputs = nn.Linear(config.hidden_size, config.num_labels)
        self.qa_outputs = LayoutLMv3ClassificationHead(config, pool_feature=False)

        self.init_weights()

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        valid_span=None,
        head_mask=None,
        inputs_embeds=None,
        start_positions=None,
        end_positions=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
        bbox=None,
        images=None,
    ):
        r"""
        start_positions (:obj:`torch.LongTensor` of shape :obj:`(batch_size,)`, `optional`):
            Labels for position (index) of the start of the labelled span for computing the token classification loss.
            Positions are clamped to the length of the sequence (:obj:`sequence_length`). Position outside of the
            sequence are not taken into account for computing the loss.
        end_positions (:obj:`torch.LongTensor` of shape :obj:`(batch_size,)`, `optional`):
            Labels for position (index) of the end of the labelled span for computing the token classification loss.
            Positions are clamped to the length of the sequence (:obj:`sequence_length`). Position outside of the
            sequence are not taken into account for computing the loss.
        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        outputs = self.layoutlmv3(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            bbox=bbox,
            images=images,
            valid_span=valid_span,
        )

        sequence_output = outputs[0]

        logits = self.qa_outputs(sequence_output)
        start_logits, end_logits = logits.split(1, dim=-1)
        start_logits = start_logits.squeeze(-1).contiguous()
        end_logits = end_logits.squeeze(-1).contiguous()

        total_loss = None
        if start_positions is not None and end_positions is not None:
            # If we are on multi-GPU, split add a dimension
            if len(start_positions.size()) > 1:
                start_positions = start_positions.squeeze(-1)
            if len(end_positions.size()) > 1:
                end_positions = end_positions.squeeze(-1)
            # sometimes the start/end positions are outside our model inputs, we ignore these terms
            ignored_index = start_logits.size(1)
            start_positions = start_positions.clamp(0, ignored_index)
            end_positions = end_positions.clamp(0, ignored_index)

            loss_fct = CrossEntropyLoss(ignore_index=ignored_index)
            start_loss = loss_fct(start_logits, start_positions)
            end_loss = loss_fct(end_logits, end_positions)
            total_loss = (start_loss + end_loss) / 2

        if not return_dict:
            output = (start_logits, end_logits) + outputs[2:]
            return ((total_loss,) + output) if total_loss is not None else output

        return QuestionAnsweringModelOutput(
            loss=total_loss,
            start_logits=start_logits,
            end_logits=end_logits,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )


class LayoutLMv3ForSequenceClassification(LayoutLMv3PreTrainedModel):
    _keys_to_ignore_on_load_missing = [r"position_ids"]

    def __init__(self, config):
        super().__init__(config)
        self.num_labels = config.num_labels
        self.config = config
        self.layoutlmv3 = LayoutLMv3Model(config)
        self.classifier = LayoutLMv3ClassificationHead(config, pool_feature=False)

        self.init_weights()

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        valid_span=None,
        head_mask=None,
        inputs_embeds=None,
        labels=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
        bbox=None,
        images=None,
    ):
        r"""
        labels (:obj:`torch.LongTensor` of shape :obj:`(batch_size,)`, `optional`):
            Labels for computing the sequence classification/regression loss. Indices should be in :obj:`[0, ...,
            config.num_labels - 1]`. If :obj:`config.num_labels == 1` a regression loss is computed (Mean-Square loss),
            If :obj:`config.num_labels > 1` a classification loss is computed (Cross-Entropy).
        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        outputs = self.layoutlmv3(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            bbox=bbox,
            images=images,
            valid_span=valid_span,
        )

        sequence_output = outputs[0][:, 0, :]
        logits = self.classifier(sequence_output)

        loss = None
        if labels is not None:
            if self.config.problem_type is None:
                if self.num_labels == 1:
                    self.config.problem_type = "regression"
                elif self.num_labels > 1 and (labels.dtype == torch.long or labels.dtype == torch.int):
                    self.config.problem_type = "single_label_classification"
                else:
                    self.config.problem_type = "multi_label_classification"

            if self.config.problem_type == "regression":
                loss_fct = MSELoss()
                if self.num_labels == 1:
                    loss = loss_fct(logits.squeeze(), labels.squeeze())
                else:
                    loss = loss_fct(logits, labels)
            elif self.config.problem_type == "single_label_classification":
                loss_fct = CrossEntropyLoss()
                loss = loss_fct(logits.view(-1, self.num_labels), labels.view(-1))
            elif self.config.problem_type == "multi_label_classification":
                loss_fct = BCEWithLogitsLoss()
                loss = loss_fct(logits, labels)

        if not return_dict:
            output = (logits,) + outputs[2:]
            return ((loss,) + output) if loss is not None else output

        return SequenceClassifierOutput(
            loss=loss,
            logits=logits,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )



import os
import json

import torch
from torch.utils.data.dataset import Dataset
from torchvision import transforms
from PIL import Image

from .image_utils import Compose, RandomResizedCropAndInterpolationWithTwoPic

XFund_label2ids = {
    "O":0,
    'B-HEADER':1,
    'I-HEADER':2,
    'B-QUESTION':3,
    'I-QUESTION':4,
    'B-ANSWER':5,
    'I-ANSWER':6,
}

class xfund_dataset(Dataset):
    def box_norm(self, box, width, height):
        def clip(min_num, num, max_num):
            return min(max(num, min_num), max_num)

        x0, y0, x1, y1 = box
        x0 = clip(0, int((x0 / width) * 1000), 1000)
        y0 = clip(0, int((y0 / height) * 1000), 1000)
        x1 = clip(0, int((x1 / width) * 1000), 1000)
        y1 = clip(0, int((y1 / height) * 1000), 1000)
        assert x1 >= x0
        assert y1 >= y0
        return [x0, y0, x1, y1]

    def get_segment_ids(self, bboxs):
        segment_ids = []
        for i in range(len(bboxs)):
            if i == 0:
                segment_ids.append(0)
            else:
                if bboxs[i - 1] == bboxs[i]:
                    segment_ids.append(segment_ids[-1])
                else:
                    segment_ids.append(segment_ids[-1] + 1)
        return segment_ids

    def get_position_ids(self, segment_ids):
        position_ids = []
        for i in range(len(segment_ids)):
            if i == 0:
                position_ids.append(2)
            else:
                if segment_ids[i] == segment_ids[i - 1]:
                    position_ids.append(position_ids[-1] + 1)
                else:
                    position_ids.append(2)
        return position_ids

    def load_data(
            self,
            data_file,
    ):
        # re-org data format
        total_data = {"id": [], "lines": [], "bboxes": [], "ner_tags": [], "image_path": []}
        for i in range(len(data_file['documents'])):
            width, height = data_file['documents'][i]['img']['width'], data_file['documents'][i]['img'][
                'height']

            cur_doc_lines, cur_doc_bboxes, cur_doc_ner_tags, cur_doc_image_path = [], [], [], []
            for j in range(len(data_file['documents'][i]['document'])):
                cur_item = data_file['documents'][i]['document'][j]
                cur_doc_lines.append(cur_item['text'])
                cur_doc_bboxes.append(self.box_norm(cur_item['box'], width=width, height=height))
                cur_doc_ner_tags.append(cur_item['label'])
            total_data['id'] += [len(total_data['id'])]
            total_data['lines'] += [cur_doc_lines]
            total_data['bboxes'] += [cur_doc_bboxes]
            total_data['ner_tags'] += [cur_doc_ner_tags]
            total_data['image_path'] += [data_file['documents'][i]['img']['fname']]

        # tokenize text and get bbox/label
        total_input_ids, total_bboxs, total_label_ids = [], [], []
        for i in range(len(total_data['lines'])):
            cur_doc_input_ids, cur_doc_bboxs, cur_doc_labels = [], [], []
            for j in range(len(total_data['lines'][i])):
                cur_input_ids = self.tokenizer(total_data['lines'][i][j], truncation=False, add_special_tokens=False, return_attention_mask=False)['input_ids']
                if len(cur_input_ids) == 0: continue

                cur_label = total_data['ner_tags'][i][j].upper()
                if cur_label == 'OTHER':
                    cur_labels = ["O"] * len(cur_input_ids)
                    for k in range(len(cur_labels)):
                        cur_labels[k] = self.label2ids[cur_labels[k]]
                else:
                    cur_labels = [cur_label] * len(cur_input_ids)
                    cur_labels[0] = self.label2ids['B-' + cur_labels[0]]
                    for k in range(1, len(cur_labels)):
                        cur_labels[k] = self.label2ids['I-' + cur_labels[k]]
                assert len(cur_input_ids) == len([total_data['bboxes'][i][j]] * len(cur_input_ids)) == len(cur_labels)
                cur_doc_input_ids += cur_input_ids
                cur_doc_bboxs += [total_data['bboxes'][i][j]] * len(cur_input_ids)
                cur_doc_labels += cur_labels
            assert len(cur_doc_input_ids) == len(cur_doc_bboxs) == len(cur_doc_labels)
            assert len(cur_doc_input_ids) > 0

            total_input_ids.append(cur_doc_input_ids)
            total_bboxs.append(cur_doc_bboxs)
            total_label_ids.append(cur_doc_labels)
        assert len(total_input_ids) == len(total_bboxs) == len(total_label_ids)

        # split text to several slices because of over-length
        input_ids, bboxs, labels = [], [], []
        segment_ids, position_ids = [], []
        image_path = []
        for i in range(len(total_input_ids)):
            start = 0
            cur_iter = 0
            while start < len(total_input_ids[i]):
                end = min(start + 510, len(total_input_ids[i]))

                input_ids.append([self.tokenizer.cls_token_id] + total_input_ids[i][start: end] + [self.tokenizer.sep_token_id])
                bboxs.append([[0, 0, 0, 0]] + total_bboxs[i][start: end] + [[1000, 1000, 1000, 1000]])
                labels.append([-100] + total_label_ids[i][start: end] + [-100])

                cur_segment_ids = self.get_segment_ids(bboxs[-1])
                cur_position_ids = self.get_position_ids(cur_segment_ids)
                segment_ids.append(cur_segment_ids)
                position_ids.append(cur_position_ids)
                image_path.append(os.path.join(self.args.data_dir, "images", total_data['image_path'][i]))

                start = end
                cur_iter += 1

        assert len(input_ids) == len(bboxs) == len(labels) == len(segment_ids) == len(position_ids)
        assert len(segment_ids) == len(image_path)

        res = {
            'input_ids': input_ids,
            'bbox': bboxs,
            'labels': labels,
            'segment_ids': segment_ids,
            'position_ids': position_ids,
            'image_path': image_path,
        }
        return res

    def __init__(
            self,
            args,
            tokenizer,
            mode
    ):
        self.args = args
        self.mode = mode
        self.cur_la = args.language
        self.tokenizer = tokenizer
        self.label2ids = XFund_label2ids


        self.common_transform = Compose([
            RandomResizedCropAndInterpolationWithTwoPic(
                size=args.input_size, interpolation=args.train_interpolation,
            ),
        ])

        self.patch_transform = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize(
                mean=torch.tensor((0.5, 0.5, 0.5)),
                std=torch.tensor((0.5, 0.5, 0.5)))
        ])

        data_file = json.load(
            open(os.path.join(args.data_dir, "{}.{}.json".format(self.cur_la, 'train' if mode == 'train' else 'val')),
                 'r'))

        self.feature = self.load_data(data_file)

    def __len__(self):
        return len(self.feature['input_ids'])

    def __getitem__(self, index):
        input_ids = self.feature["input_ids"][index]

        # attention_mask = self.feature["attention_mask"][index]
        attention_mask = [1] * len(input_ids)
        labels = self.feature["labels"][index]
        bbox = self.feature["bbox"][index]
        segment_ids = self.feature['segment_ids'][index]
        position_ids = self.feature['position_ids'][index]

        img = pil_loader(self.feature['image_path'][index])
        for_patches, _ = self.common_transform(img, augmentation=False)
        patch = self.patch_transform(for_patches)

        assert len(input_ids) == len(attention_mask) == len(labels) == len(bbox) == len(segment_ids)

        res = {
            "input_ids": input_ids,
            "attention_mask": attention_mask,
            "labels": labels,
            "bbox": bbox,
            "segment_ids": segment_ids,
            "position_ids": position_ids,
            "images": patch,
        }
        return res

def pil_loader(path: str) -> Image.Image:
    # open path as file to avoid ResourceWarning (https://github.com/python-pillow/Pillow/issues/835)
    with open(path, 'rb') as f:
        img = Image.open(f)
        return img.convert('RGB')


'''
Author: dbliu shaxunyeman@gmail.com
Date: 2024-09-02 17:55:19
LastEditors: dbliu shaxunyeman@gmail.com
LastEditTime: 2024-09-05 09:02:47
FilePath: /magic-doc/magic_doc/conv/pdf_pp_structurev2.py
Description: 
'''
from pathlib import Path
import os

from magic_pdf.dict2md.ocr_mkcontent import union_make
from magic_pdf.libs.MakeContentConfig import MakeMode, DropMode
from magic_pdf.libs.json_compressor import JsonCompressor
from magic_pdf.model.doc_analyze_by_custom_model import doc_analyze
from magic_pdf.rw.AbsReaderWriter import AbsReaderWriter
from magic_pdf.pipe.UNIPipe import UNIPipe

from loguru import logger

from magic_doc.conv.base import BaseConv
from magic_doc.progress.filepupdator import FileBaseProgressUpdator
from magic_doc.progress.pupdator import ConvProgressUpdator
from magic_doc.utils.null_writer import NullWriter
from magic_doc.common.default_config import DEFAULT_CONFIG

NULL_IMG_DIR = "/tmp"

class SingletonModelWrapper:

    def __new__(cls):
        if not hasattr(cls, "instance"):
            from magic_doc.model.doc_analysis_by_pp import PaddleDocAnalysis
            cls.instance = super(SingletonModelWrapper, cls).__new__(cls)
            cls.instance.model = PaddleDocAnalysis(model_load_on_each_gpu_count=int(DEFAULT_CONFIG["pdf"]["fast"]["liteocrmodelinstance"]))
        return cls.instance
    
    def __call__(self, bytes: bytes):
        from magic_pdf.model.doc_analyze_by_custom_model import load_images_from_pdf
        images = load_images_from_pdf(bytes, dpi=200)
        return self.model(images) # type: ignore


class Pdf(BaseConv):
    def to_md(self, bits: bytes | str, pupdator: ConvProgressUpdator, **kwargs) -> str:
        model = SingletonModelWrapper()
        model_list = model(bits)
        pupdator.update(50)
        jso_useful_key = {
            "_pdf_type": "ocr",
            "model_list": model_list,
        }
        image_writer = NullWriter()
        pipe = UNIPipe(bits, jso_useful_key, image_writer, is_debug=True)  # type: ignore
        # pipe.pipe_classify() # 默认ocrpipe的时候不需要再做分类，可以节省时间
        pipe.pipe_parse()
        pupdator.update(100)

        pdf_mid_data = JsonCompressor.decompress_json(pipe.get_compress_pdf_mid_data())
        pdf_info_list = pdf_mid_data["pdf_info"]
        md_content = union_make(pdf_info_list, MakeMode.NLP_MD, DropMode.NONE, NULL_IMG_DIR)
        return md_content  # type: ignore

    def to_mid_result(self, image_writer: AbsReaderWriter, bits: bytes | str, pupdator: ConvProgressUpdator) \
            -> list[dict] | dict:
        model = SingletonModelWrapper()
        pupdator.update(0)
        model_list = model(bits)
        pupdator.update(50)
        jso_useful_key = {
            "_pdf_type": "ocr",
            "model_list": model_list,
        }

        pipe = UNIPipe(bits, jso_useful_key, image_writer, is_debug=True)  # type: ignore
        # pipe.pipe_classify()
        pipe.pipe_parse()
        pupdator.update(100)

        pdf_mid_data = JsonCompressor.decompress_json(pipe.get_compress_pdf_mid_data())
        pdf_info_list = pdf_mid_data["pdf_info"]
        return pdf_info_list


if __name__ == "__main__":
    pupdator = FileBaseProgressUpdator("/tmp/p.txt")
    pdf = Pdf()
    logger.info(
        pdf.to_md(Path(r"D:\project\20231108code-clean\linshixuqiu\pdf_dev\新模型\j.sna.2004.11.030.pdf").read_bytes(), pupdator))



'''
Author: dbliu shaxunyeman@gmail.com
Date: 2024-09-02 17:55:19
LastEditors: dbliu shaxunyeman@gmail.com
LastEditTime: 2024-09-05 09:02:21
FilePath: /magic-doc/magic_doc/conv/doc_libreoffice.py
Description: 
'''
import os
import tempfile
from pathlib import Path
from subprocess import Popen

from loguru import logger

from magic_doc.contrib.model import Page
from magic_doc.contrib.office.docx_extract import DocxExtractor
from magic_doc.conv.base import BaseConv
from magic_doc.progress.filepupdator import FileBaseProgressUpdator
from magic_doc.progress.pupdator import ConvProgressUpdator


class Doc(BaseConv):

    def __init__(self):
        super().__init__()

    def to_md(self, bits: bytes, pupdator: ConvProgressUpdator, **kwargs) -> str:
        page_list = self.doc_to_pagelist(bits, pupdator)
        md_content_list = []
        for page in page_list:
            page_content_list = page["content_list"]
            total = len(page_content_list)
            for index, content in enumerate(page_content_list):
                progress = 50 + int(index / total * 50)
                # logger.info(f"progress: {progress}")
                pupdator.update(progress)
                if content["type"] == "image":
                    pass
                elif content["type"] in ["text", "md"]:
                    data = content["data"]
                    md_content_list.append(data)
        return "\n".join(md_content_list)

    def doc_to_docx(self, doc_path: str, dir_path: str) -> str:
        cmd = f'soffice --headless --convert-to docx "{doc_path}" --outdir "{dir_path}"'
        logger.info(cmd)
        process = Popen(cmd, shell=True)
        process.wait()
        fname = str(Path(doc_path).stem)
        docx_path = os.path.join(os.path.dirname(doc_path), f"{fname}.docx")
        if not os.path.exists(docx_path):
            # logger.error(f"> !!! File conversion failed {doc_path} ==> {docx_path}")
            raise Exception(f"> !!! File conversion failed {doc_path} ==> {docx_path}")
        else:
            return docx_path

    def doc_to_pagelist(self, bits, pupdator: ConvProgressUpdator) -> list[Page]:
        with tempfile.TemporaryDirectory() as temp_path:
            temp_dir = Path(temp_path)
            media_dir = temp_dir / "media"
            media_dir.mkdir()
            file_path = temp_dir / "tmp.doc"
            file_path.write_bytes(bits)
            docx_file_path = self.doc_to_docx(str(file_path), str(temp_path))
            pupdator.update(50)
            docx_extractor = DocxExtractor()
            pages = docx_extractor.extract(
                Path(docx_file_path), "tmp", temp_dir, media_dir, True
            )
            pupdator.update(80)
            return pages


if __name__ == "__main__":
    pupdator = FileBaseProgressUpdator("/tmp/p.txt")
    doc = Doc()
    if 0:
        logger.info(
            doc.to_md(
                Path(
                    r"D:\project\20240514magic_doc\doc_ppt\doc\demo\文本+表+图1.doc"
                ).read_bytes(),
                pupdator,
            )
        )
    if 1:
        print(
            doc.to_md(
                Path(
                    r"/opt/data/magic_doc/20240605/doc/【英文-习题】MIT_prs_w05d2.doc"
                ).read_bytes(),
                pupdator,
            )
        )



# coding=utf-8
# Copyright 2018 The Open AI Team Authors and The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fast Tokenization classes for LayoutLMv3, refer to RoBERTa."""


from transformers.models.roberta.tokenization_roberta_fast import RobertaTokenizerFast
from transformers.utils import logging

from .tokenization_layoutlmv3 import LayoutLMv3Tokenizer


logger = logging.get_logger(__name__)

VOCAB_FILES_NAMES = {"vocab_file": "vocab.json", "merges_file": "merges.txt", "tokenizer_file": "tokenizer.json"}


class LayoutLMv3TokenizerFast(RobertaTokenizerFast):
    vocab_files_names = VOCAB_FILES_NAMES
    # pretrained_vocab_files_map = PRETRAINED_VOCAB_FILES_MAP
    # max_model_input_sizes = PRETRAINED_POSITIONAL_EMBEDDINGS_SIZES
    model_input_names = ["input_ids", "attention_mask"]
    slow_tokenizer_class = LayoutLMv3Tokenizer





from magic_doc.progress.pupdator import ConvProgressUpdator


class FileBaseProgressUpdator(ConvProgressUpdator):
    def __init__(self, progress_file_path:str):
        self.__progress_file_path = progress_file_path
    
    def do_update(self, progress:int) -> bool:
        with open(self.__progress_file_path, 'w', encoding='utf-8') as fout:
            fout.write(str(int(progress)))

        return True


'''
Author: dbliu shaxunyeman@gmail.com
Date: 2024-09-02 17:55:19
LastEditors: dbliu shaxunyeman@gmail.com
LastEditTime: 2024-09-06 17:20:16
FilePath: /magic-doc/magic_doc/libs/version.py
Description: 
'''
__version__ = "0.1.5"